import os
from copy import deepcopy
from pathlib import Path

import napari
import napari.utils.notifications
import numpy as np
import tifffile as tff
from PyQt5 import QtWidgets

from fibsem.detection import detection
from fibsem.detection import utils as det_utils
from fibsem.detection.detection import DetectedFeatures
from fibsem.segmentation import model as fibsem_model
from fibsem.ui.utils import message_box_ui
from fibsem.segmentation.model import load_model
from fibsem.structures import (
    BeamType,
    FibsemImage,
    Point,
)
from fibsem.ui import _stylesheets
from PyQt5.QtCore import pyqtSignal
from fibsem.ui.qtdesigner_files import FibsemDetectionWidget
import logging

CHECKPOINT_PATH = os.path.join(os.path.dirname(fibsem_model.__file__), "models", "model4.pt")
CLASS_COLORS = {0: "black", 1: "red", 2: "green", 3: "cyan", 4: "yellow", 5: "magenta", 6: "blue"}

class FibsemDetectionWidgetUI(FibsemDetectionWidget.Ui_Form, QtWidgets.QDialog):
    continue_signal = pyqtSignal(DetectedFeatures)

    def __init__(
        self,
        viewer: napari.Viewer,
        image: FibsemImage,
        detected_features: DetectedFeatures = None,
        model: fibsem_model.SegmentationModel = None,
        _SEG_MODE: bool = True,
        _DET_MODE: bool = True,
        _EVAL_MODE: bool = False,
        parent=None,
    ):
        super(FibsemDetectionWidgetUI, self).__init__(parent=parent)
        self.setupUi(self)

        self.viewer = viewer
        self.image = image
        self.model = model
        self.image_paths = None

        self._EVAL_MODE = _EVAL_MODE
        self._USER_CORRECTED = False
        self._USE_ALL_FEATURES = False

        self.USE_SEGMENTATION = True
        self.USE_FEATURE_DETECTION = _DET_MODE
        self.USE_EVALUATION = _EVAL_MODE

        self._image_layer = None
        self._mask_layer = None
        self._features_layer = None

        self.checkBox_use_segmentation.setChecked(self.USE_SEGMENTATION)
        self.checkBox_use_segmentation.setVisible(False)
        self.checkBox_use_feature_detection.setChecked(self.USE_FEATURE_DETECTION)
        self.checkBox_use_evaluation.setChecked(self.USE_EVALUATION)
        


        self.setup_connections()

        # set detected features
        if detected_features is not None:
            self.set_detected_features(detected_features)

        # message_box_ui(title="Feature Detection Validation",text="Ensure that the features detected are located accurately, if not, please manually correct the locations by dragging. Once locations are valid, click continue",buttons=QtWidgets.QMessageBox.Ok)

    def setup_connections(self):
        self.label_instructions.setText(
            """Drag the detected feature positions to move them. Press continue when finished."""
        )
        self.pushButton_continue.clicked.connect(self.continue_button_clicked)

        self.pushButton_run_feature_detection.clicked.connect(
            self.run_feature_detection
        )
        self.pushButton_run_feature_detection.setStyleSheet(_stylesheets._GREEN_PUSHBUTTON_STYLE)
        # ui
        # self.checkBox_use_segmentation.stateChanged.connect(self._toggle_ui)
        self.checkBox_use_feature_detection.stateChanged.connect(self._toggle_ui)
        self.checkBox_use_evaluation.stateChanged.connect(self._toggle_ui)
        self.checkBox_move_features.stateChanged.connect(self._toggle_ui)
        self.checkBox_move_features.setEnabled(False)  # disabled until features are loaded
        self.checkBox_all_features.stateChanged.connect(self._toggle_ui)
        self.checkBox_show_info.stateChanged.connect(self._toggle_ui)

        # features
        self.comboBox_feature_1.addItems([f.name for f in detection.__FEATURES__])
        self.comboBox_feature_2.addItems([f.name for f in detection.__FEATURES__])

        # model
        self.pushButton_load_model.clicked.connect(self.load_model)
        self.pushButton_load_model.setStyleSheet(_stylesheets._GREEN_PUSHBUTTON_STYLE)
        self.lineEdit_encoder.setText("resnet34")
        self.lineEdit_checkpoint.setText(CHECKPOINT_PATH)
        self.spinBox_num_classes.setValue(3)

        self.continue_signal.connect(self.do_something)

        self.lineEdit_image_path_folder.setText(f"/home/patrick/github/data/liftout/training/train/images")
        self.label_image_path_num.setText("")

        self.label_header_evalulation.setVisible(self._EVAL_MODE)
        self.lineEdit_image_path_folder.setVisible(self._EVAL_MODE)
        self.pushButton_load_images.setVisible(self._EVAL_MODE)
        self.pushButton_load_images.clicked.connect(self.load_image_folder)
        self.pushButton_load_images.setStyleSheet(_stylesheets._BLUE_PUSHBUTTON_STYLE)
        self.pushButton_previous_image.clicked.connect(self.update_image)
        self.pushButton_next_image.clicked.connect(self.update_image)
        self.pushButton_previous_image.setVisible(self._EVAL_MODE)
        self.pushButton_next_image.setVisible(self._EVAL_MODE)
        self.label_image_path_num.setVisible(self._EVAL_MODE)
        self.pushButton_next_image.setVisible(False)
        self.pushButton_previous_image.setVisible(False)
        self.label_image_path_num.setVisible(False)

        # TODO: add button to go to idx

    def load_image_folder(self):
        folder = self.lineEdit_image_path_folder.text()
        self.image_paths = sorted(list(Path(folder).glob("*.tif")))
        self.idx = 0
        self.update_image()
        self.pushButton_next_image.setVisible(True)
        self.pushButton_previous_image.setVisible(True)
        self.label_image_path_num.setVisible(True)
    
    def save_data(self):

        
        # get the updated mask
        self.detected_features.mask = self._mask_layer.data.astype(np.uint8) # type: ignore
        
        # save current data
        # fname = os.path.basename(self.image_paths[self.idx])
        det_utils.save_data(det = self.detected_features, corrected=self._USER_CORRECTED, fname=None)


    def update_image(self):

        if self.image_paths is None:
            return

        # save current data
        self.save_data()

        # update image view
        if self.sender() == self.pushButton_previous_image:
            self.idx -= 1
        if self.sender() == self.pushButton_next_image:
            self.idx += 1

        self.idx = np.clip(self.idx, 0, len(self.image_paths)-1)
        self.label_image_path_num.setText(f"{self.idx+1}/{len(self.image_paths)}")

        self.image = FibsemImage.load(self.image_paths[self.idx])

        self.run_feature_detection()

    def do_something(self, msg):
        print("do_something called")
        # print(msg)
        

    def _toggle_ui(self):

        # self.USE_SEGMENTATION = self.checkBox_use_segmentation.isChecked()

        # if self.USE_SEGMENTATION is False:
            # self.checkBox_use_feature_detection.setChecked(False)

        self.USE_FEATURE_DETECTION = self.checkBox_use_feature_detection.isChecked()
        self.USE_EVALUATION = self.checkBox_use_evaluation.isChecked()

        # # hide segmentation if not used
        # self.label_header_model.setVisible(self.USE_SEGMENTATION)
        # self.label_encoder.setVisible(self.USE_SEGMENTATION)
        # self.lineEdit_encoder.setVisible(self.USE_SEGMENTATION)
        # self.label_checkpoint.setVisible(self.USE_SEGMENTATION)
        # self.lineEdit_checkpoint.setVisible(self.USE_SEGMENTATION)
        # self.label_num_classes.setVisible(self.USE_SEGMENTATION)
        # self.spinBox_num_classes.setVisible(self.USE_SEGMENTATION)
        # self.pushButton_load_model.setVisible(self.USE_SEGMENTATION)

        # hide feature detection if not used
        self.label_header_features.setVisible(self.USE_FEATURE_DETECTION)
        self.label_feature.setVisible(self.USE_FEATURE_DETECTION)
        self.comboBox_feature_1.setVisible(self.USE_FEATURE_DETECTION)
        self.comboBox_feature_2.setVisible(self.USE_FEATURE_DETECTION)
        self.checkBox_all_features.setVisible(self.USE_FEATURE_DETECTION)
        self.checkBox_move_features.setVisible(self.USE_FEATURE_DETECTION)
        self.checkBox_show_info.setVisible(self.USE_FEATURE_DETECTION)
        # self.pushButton_run_feature_detection.setVisible(self.USE_FEATURE_DETECTION)


        # hide evaluation if not used
        self.label_header_evalulation.setVisible(self.USE_EVALUATION)
        self.label_dataset.setVisible(self.USE_EVALUATION)
        self.lineEdit_image_path_folder.setVisible(self.USE_EVALUATION)
        self.pushButton_load_images.setVisible(self.USE_EVALUATION)
        self.pushButton_previous_image.setVisible(self.USE_EVALUATION)
        self.pushButton_next_image.setVisible(self.USE_EVALUATION)
        self.label_image_path_num.setVisible(self.USE_EVALUATION)

        if self.USE_FEATURE_DETECTION:
            self.pushButton_run_feature_detection.setText("Run Feature Detection")
        else:
            self.pushButton_run_feature_detection.setText("Run Segmentation")

    

        # features
        if not self.USE_FEATURE_DETECTION:
            if "features" in self.viewer.layers:
                self.viewer.layers.remove(self._features_layer)
                self._features_layer = None

            return
        
        self._USE_ALL_FEATURES = self.checkBox_all_features.isChecked()
        self.label_feature.setVisible(not self._USE_ALL_FEATURES)
        self.comboBox_feature_1.setVisible(not self._USE_ALL_FEATURES)
        self.comboBox_feature_2.setVisible(not self._USE_ALL_FEATURES)
        
        # info
        self.label_info.setVisible(self.checkBox_show_info.isChecked())
        
        if "features" in self.viewer.layers:
            self.checkBox_move_features.setEnabled(True)
            
            if self.checkBox_move_features.isChecked():
                self.viewer.layers.selection.active = self._features_layer
            else:
                self.viewer.layers.selection.active = self._image_layer

    def run_feature_detection(self):

        self._USER_CORRECTED = False # reset user corrected flag

        if self.model is None:
            self.load_model()

        if self._USE_ALL_FEATURES or self.USE_FEATURE_DETECTION is False:
            features = [f() for f in detection.__FEATURES__]
        else:
            features = (
                detection.__FEATURES__[self.comboBox_feature_1.currentIndex()](),
                detection.__FEATURES__[self.comboBox_feature_2.currentIndex()](),
            )

        # detect features
        pixelsize = 25e-9 # TODO: get from metadata
        det = detection.detect_features(

            deepcopy(self.image.data), self.model, 
            features=features, 
            pixelsize=pixelsize,
            filter=True, point=None
        )

        self.set_detected_features(det)

    def load_model(self):
        checkpoint = self.lineEdit_checkpoint.text()
        encoder = self.lineEdit_encoder.text()
        num_classes = self.spinBox_num_classes.value()
        logging.info(f"loading checkpoint: {checkpoint}...")
        self.model = load_model(checkpoint=checkpoint, encoder=encoder, nc=num_classes)
        napari.utils.notifications.show_info(
            f"Model loaded: {os.path.basename(checkpoint)}"
        )

    def continue_button_clicked(self):
        
        # save current data
        self.save_data()

        # emit signal
        self.continue_signal.emit(self.detected_features)
        print("continue signal emitted")

        

        if not self._EVAL_MODE:
            self.close()
            self.viewer.close()

    def set_detected_features(self, det_features: DetectedFeatures):
        self.detected_features = det_features

        self.update_features_ui()

    def update_features_ui(self):
        
        # update combo box
        if not self._USE_ALL_FEATURES:
            self.comboBox_feature_1.setCurrentText(self.detected_features.features[0].name)
            self.comboBox_feature_2.setCurrentText(self.detected_features.features[1].name)

        try:
            self._image_layer.data = self.detected_features.image
        except:
            self._image_layer = self.viewer.add_image(
                self.detected_features.image, name="image", opacity=0.7, blending="additive",
            )

        # add mask to viewer
        try:
            self._mask_layer.data = self.detected_features.mask
        except:
            self._mask_layer = self.viewer.add_labels(self.detected_features.mask, 
                                                      name="mask", 
                                                      opacity=0.7,
                                                      blending="additive", 
                                                      color=CLASS_COLORS)

        # if the features layer already exists, remove the layer
        if "features" in self.viewer.layers:
            self.viewer.layers.remove(self._features_layer)
            self._features_layer = None

        if not self.USE_FEATURE_DETECTION:
            napari.utils.notifications.show_info(f"Segmentation finished.")
            return

        # add points to viewer
        data = []
        for feature in self.detected_features.features:
            x, y = feature.px
            data.append([y, x])


        text = {
            "string": [feature.name for feature in self.detected_features.features],
            "color": "white",
            "translation": np.array([-30, 0]),
        }

        self._features_layer = self.viewer.add_points(
            data,
            name="features",
            text=text,
            size=20,
            edge_width=7,
            edge_width_is_relative=False,
            edge_color="transparent",
            face_color=[feature.color for feature in self.detected_features.features],
            blending="translucent",
        )

        # set points layer to select mode and active
        self._features_layer.mode = "select"

        # when the point is moved update the feature
        self._features_layer.events.data.connect(self.update_point)

        self.update_info()
        self.checkBox_move_features.setChecked(True)
        self._toggle_ui()

        napari.utils.notifications.show_info(f"Features Detected")

    def update_info(self):
        
        if len(self.detected_features.features) != 2:
            self.label_info.setText("Please select two features for info.")
            return
        
        self.label_info.setText(
            f"""Moving {self.detected_features.features[0].name} to {self.detected_features.features[1].name}
        \n{self.detected_features.features[0].name}: {self.detected_features.features[0].px}
        \n{self.detected_features.features[1].name}: {self.detected_features.features[1].px}
        \ndx={self.detected_features.distance.x*1e6:.2f}um, dy={self.detected_features.distance.y*1e6:.2f}um
        User Corrected: {self._USER_CORRECTED}
        """
        )

    def update_point(self, event):
        logging.info(f"{event.source.name} changed its data!")

        layer = self.viewer.layers[f"{event.source.name}"]  # type: ignore

        # get the data
        data = layer.data

        # get which point was moved
        index: list[int] = list(layer.selected_data)  
                
        if len(data) != len(self.detected_features.features):
            # loop backwards to remove the features
            for idx in index[::-1]:
                logging.info(f"point deleted: {self.detected_features.features[idx].name}")
                self.detected_features.features.pop(idx)

        else: 
            for idx in index:
                
                logging.info(f"point moved: {self.detected_features.features[idx].name} to {data[idx]}") # TODO: fix for logging statistics
                
                # update the feature
                self.detected_features.features[idx].px = Point(
                    x=data[idx][1], y=data[idx][0]
                )

        self._USER_CORRECTED = True
        self.update_info()



    def _get_detected_features(self):
        return self.detected_features


from fibsem.microscope import FibsemMicroscope
from fibsem.structures import MicroscopeSettings
from fibsem.detection.detection import Feature, DetectedFeatures
from fibsem import acquire
def detection_ui(image: FibsemImage, model: fibsem_model.SegmentationModel, features: list[Feature], validate: bool = True) -> DetectedFeatures:
    print("GOT TO DETECTION UI")
    pixelsize = image.metadata.pixel_size.x if image.metadata is not None else 25e-9

    # detect features
    det = detection.detect_features(
        deepcopy(image.data), model, features=features, pixelsize=pixelsize
    )

    if validate:
        viewer = napari.Viewer(ndisplay=2)
        det_widget_ui = FibsemDetectionWidgetUI(
            viewer=viewer, 
            image = image, 
            _EVAL_MODE=False,)
        
        det_widget_ui.set_detected_features(det)

        viewer.window.add_dock_widget(
            det_widget_ui, area="right", add_vertical_stretch=False, name="Fibsem Feature Detection"
        )
        det_widget_ui.exec_()
        # napari.run()

        det = det_widget_ui.detected_features

    
    return det



def detection_movement(microscope: FibsemMicroscope, settings: MicroscopeSettings, validate: bool = True, _load_image: bool = False) -> DetectedFeatures:

    if _load_image:
        # load image
        image = FibsemImage.load(os.path.join(os.path.dirname(detection.__file__), "test_image.tif"))
    else:
        image = acquire.new_image(microscope, settings.image)

    # TODO: read from config?
    checkpoint = str(CHECKPOINT_PATH)
    encoder="resnet34"
    num_classes = 3
    model = load_model(checkpoint=checkpoint, encoder=encoder, nc=num_classes)
 
    features = [detection.NeedleTip(), detection.LamellaCentre()]
    det = detection_ui(image, model, features, validate=validate)

    return det



def main():

    validate = True
    _load_image = True

    from fibsem import utils
    microscope, settings = utils.setup_session(manufacturer="Demo")

    det = detection_movement(microscope, settings, validate=validate, _load_image=_load_image)

    detection.move_based_on_detection(
    microscope,
    settings,
    det=det,
    beam_type=settings.image.beam_type,
)


if __name__ == "__main__":
    main()


# DONE
# - convert to use binary masks instead of rgb - DOne
# - add mask, rgb to detected features + save to file  # DONE
# - convert mask layer to label not image # DONE
# - save detected features to file on prev / save image # DONE
# - add n detections, not just two.. if no features are passed... use all?
# - add toggles for seg / feature detection / eval
# - maybe integrate as labelling ui? -> assisted labelling
# - toggle show info checkbox

# TODO:
# - convert detected features / detection to take in Union[FibsemImage, np.ndarray]
# - edittable mask -> rerun detection 
# - abstract segmentation model widget
# - need to ensure feature det is only enabled if seg is enabled
# - need seg to be enabled if feature det is enabled same for eval


import os
import sys
import win32com.client
from fibsem.config import BASE_PATH
# Get the user's desktop folder
desktop = os.path.join(os.path.expanduser('~'), 'Desktop')

# Specify the target file (e.g., a script or program you want to create a shortcut for)
target_file = os.path.join(BASE_PATH, "run_ui.bat") # Replace with your program's path

# Create a shortcut name
shortcut_name = 'FibsemUI.lnk'

# Create a shortcut on the desktop
shell = win32com.client.Dispatch("WScript.Shell")
shortcut = shell.CreateShortCut(os.path.join(desktop, shortcut_name))
shortcut.TargetPath = target_file
shortcut.save()

print(f"Shortcut to '{target_file}' created on the desktop as '{shortcut_name}'")


import time
import logging

import numpy as np

try:
    from autoscript_sdb_microscope_client import SdbMicroscopeClient
    from autoscript_sdb_microscope_client.enumerations import (
        ManipulatorCoordinateSystem,
        ManipulatorSavedPosition,
    )
    from autoscript_sdb_microscope_client.structures import (
        ManipulatorPosition,
        MoveSettings,
        StagePosition,
    )
    THERMO = True
except:
    THERMO = False

from fibsem.structures import BeamType, MicroscopeSettings
from fibsem.microscope import FibsemMicroscope

# from fibsem.detection.detection import DetectionResult, FeatureType


############################## NEEDLE ##############################

if THERMO:
    def retract_multichem(microscope: SdbMicroscopeClient) -> None:
        # Retract the multichem
        logging.info(f"retracting multichem")
        multichem = microscope.gas.get_multichem()
        multichem.retract()
        logging.info(f"retract multichem complete")

        return

def rotation_angle_is_larger(angle1: float, angle2: float, atol: float = 90) -> bool:
    """Check the rotation angles are large

    Args:
        angle1 (float): angle1 (radians)
        angle2 (float): angle2 (radians)
        atol : tolerance (degrees)

    Returns:
        bool: rotation angle is larger than atol
    """

    return angle_difference(angle1, angle2) > (np.deg2rad(atol))


def rotation_angle_is_smaller(angle1: float, angle2: float, atol: float = 5) -> bool:
    """Check the rotation angles are large

    Args:
        angle1 (float): angle1 (radians)
        angle2 (float): angle2 (radians)
        atol : tolerance (degrees)

    Returns:
        bool: rotation angle is smaller than atol
    """

    return angle_difference(angle1, angle2) < (np.deg2rad(atol))


def angle_difference(angle1: float, angle2: float) -> float:
    """Return the difference between two angles, accounting for greater than 360, less than 0 angles

    Args:
        angle1 (float): angle1 (radians)
        angle2 (float): angle2 (radians)

    Returns:
        float: _description_
    """
    angle1 %= 2 * np.pi
    angle2 %= 2 * np.pi

    large_angle = np.max([angle1, angle2])
    small_angle = np.min([angle1, angle2])

    return min((large_angle - small_angle), ((2 * np.pi + small_angle - large_angle)))



import napari

import logging

import fibsem
from fibsem.ui.qtdesigner_files import FibsemFeatureDetectionUI as FibsemFeatureLabellingUI
from fibsem.ui.utils import (
    _get_directory_ui,
    _get_file_ui,
    _get_text_ui,
    _get_save_file_ui,
)

from PyQt5 import QtWidgets
from PyQt5.QtCore import pyqtSignal
import glob
import os

import napari.utils.notifications
from fibsem.detection.detection import __FEATURES__, get_feature

from napari.layers import Points
import numpy as np
import os
import glob
import pandas as pd

from typing import Optional, Any
import copy
from fibsem.structures import Point
from fibsem.ui import _stylesheets

logging.basicConfig(level=logging.INFO)


def create_dataframe(filenames, method="null"):
    df = pd.DataFrame(
        columns=[
            "filename",
            "feature",
            "px.x",
            "px.y",
            "pixelsize",
            "corrected",
            "method",
            "experiment",
        ]
    )
    df["filename"] = filenames
    df["filename"] = df["filename"].apply(lambda x: os.path.basename(x))
    df["path"] = df["filename"].apply(lambda x: os.path.dirname(x))
    df["method"] = method
    return df


# TODO: remove this debug path
TEST_PATH_DEBUG = "/home/patrick/github/data/autolamella-paper/model-development/train/serial-liftout/test"

# new line char in html
NL = "<br>"

DATA_LOADED_INSTRUCTIONS = f"""
<h3>Instructions</h3>
Use the Slider to Change Images{NL}{NL}

<strong>Add Mode</strong>{NL}
Select Feature from Dropdown{NL}
Left Click to Add Feature{NL}{NL}

<strong>Edit Mode</strong>{NL}
Press Edit to enter Edit Mode{NL}
Select and Drag the Feature to move it.{NL}
Select a Feature and pPress Delete Key{NL}{NL}

<strong>HotKeys</strong>{NL}
Press 2 to enter Add Mode{NL}
Press 3 to enter Edit Mode{NL}
Press 4 to enter Pan/Zoom Mode{NL}
Press Delete in Edit Mode to Delete Feature{NL}{NL}

<strong>Save</strong>{NL}
Data is automatically saved to csv file when a feature is added, edited or deleted.
"""
START_INSTRUCTIONS = f"""
<h3>Instructions</h3>
Load Data from File Menu{NL}{NL}

<strong>Image Directory</strong>{NL}
If starting from scratch, load a directory of images and a csv file will be created for you.
Features will be saved to the csv file.{NL}{NL}

<strong>CSV File</strong>{NL}
If you are resuming work, load the csv file and the images will be loaded for you.
Features will be reloaded from the csv file.{NL}{NL}
"""

INSTRUCTIONS = {"START": START_INSTRUCTIONS, "DATA_LOADED": DATA_LOADED_INSTRUCTIONS}

FEATURES_TO_IGNORE = ["ImageCentre", "CoreFeature"]
USABLE_FEATURES = sorted([f.name for f in __FEATURES__ if f.name not in FEATURES_TO_IGNORE])

CONFIGURATION = {
    
    "FEATURES": USABLE_FEATURES,

    "UI": {
        "IMAGE_LAYER_NAME": "Image",
        "POINT_LAYER_NAME": "Features",
        "SHAPE_LAYER_NAME": "Bounding Boxes",
        "POINT_SIZE": 5,
        "TEXT_COLOR": "white",
        "TEXT_TRANSLATION": np.array([-30, 0]),
        "FACE_COLOR": "white",
        "ADD_BUTTON_COLOR": _stylesheets._GREEN_PUSHBUTTON_STYLE,
        "EDIT_BUTTON_COLOR": _stylesheets._BLUE_PUSHBUTTON_STYLE,
    },

}


# TODO:
# add support for custom features
# import features from txt file
# import features from yaml file (with color map)


def get_feature_color(feature: str) -> str:
    """try to get the feature color 
    from the feature name, with fallback 
    to base colour"""

    try:
        return get_feature(feature).color # only for fibsem features
    except:
        return CONFIGURATION["UI"]["FACE_COLOR"]


class FibsemFeatureLabellingUI(FibsemFeatureLabellingUI.Ui_MainWindow, QtWidgets.QMainWindow
):
    _minimap_signal = pyqtSignal(object)

    def __init__(self, viewer: napari.Viewer):
        super(FibsemFeatureLabellingUI, self).__init__()
        self.setupUi(self)

        self.viewer = viewer
        self.img_layer = None
        self.pts_layer = None

        self.feature = None
        self.features: list[dict] = []

        self.setup_connections()

    def setup_connections(self):
        self.actionLoad_Image_Directory.triggered.connect(self.load_image_directory)
        self.actionLoad_CSV.triggered.connect(self.load_csv)

        # controls
        self.spinBox_data_progress.valueChanged.connect(self._change_image)
        self.spinBox_data_progress.setKeyboardTracking(False)

        self.comboBox_features.addItems(CONFIGURATION["FEATURES"])
        self.comboBox_features.currentIndexChanged.connect(self._get_feature)
        self.pushButton_add_feature.clicked.connect(self._add_feature_mode)
        self.pushButton_edit_feature.clicked.connect(self._edit_feature_mode)

        self.pushButton_add_feature.setStyleSheet(CONFIGURATION["UI"]["ADD_BUTTON_COLOR"])
        self.pushButton_edit_feature.setStyleSheet(CONFIGURATION["UI"]["EDIT_BUTTON_COLOR"])

        self.update_ui_elements()

    def update_ui_elements(self):
        _data_loaded = self.img_layer is not None

        self.label_features_header.setVisible(_data_loaded)
        self.pushButton_add_feature.setVisible(_data_loaded)
        self.pushButton_edit_feature.setVisible(_data_loaded)
        self.comboBox_features.setVisible(_data_loaded)
        self.label_data_progress.setVisible(_data_loaded)
        self.spinBox_data_progress.setVisible(_data_loaded)

        self.spinBox_data_progress.setEnabled(_data_loaded)

        if _data_loaded:
            self.label_data_progress.setText(
                f"Current Image (Total: {len(self.filenames)})"
            )
            self.spinBox_data_progress.setMaximum(len(self.filenames) - 1)
            self.spinBox_data_progress.setValue(0)

            instructions = INSTRUCTIONS["DATA_LOADED"]
        else:
            instructions = INSTRUCTIONS["START"]
        self.label_instructions.setText(instructions)

    def load_image_directory(self):
        """Load a directory of images into napari"""
        print(f"Loading image directory")

        path = _get_directory_ui(
            msg="Select image directory (*.tif)", parent=self, path=TEST_PATH_DEBUG
        )

        if path is None:
            return

        filenames = sorted(glob.glob(os.path.join(path, "*.tif*")))

        if len(filenames) == 0:
            napari.utils.notifications.show_warning(
                f"No images found in {path}. Exiting"
            )
            return

        method, ret = _get_text_ui(
            msg="Enter method name",
            title="Dataset Method",
            default="autolamella",
            parent=self,
        )

        if ret == False:
            napari.utils.notifications.show_warning(f"No method name entered. Exiting.")
            return

        # create dataframe
        df = create_dataframe(filenames, method=method)

        # save dataframe
        csv_path = _get_save_file_ui(
            msg="Save CSV file",
            path=os.path.join(path, "data.csv"),
            _filter="CSV (*.csv)",
            parent=self,
        )
        if csv_path is None:
            napari.utils.notifications.show_warning(f"No CSV file selected. Exiting.")
            return

        df.to_csv(csv_path, index=False)

        # load data
        self.load_data(df, filenames, path, csv_path)

    def load_csv(self):
        """Load a CSV file into napari"""

        # get csv path from user
        csv_path = _get_file_ui(
            msg="Select CSV file",
            _filter="CSV (*.csv)",
            path=TEST_PATH_DEBUG,
            parent=self,
        )

        if csv_path is None:
            napari.utils.notifications.show_warning(f"No CSV file selected. Exiting.")
            return

        df = pd.read_csv(csv_path)

        # sort on filenames
        df.sort_values(by=["filename"], inplace=True)

        # get image path from user
        path = _get_directory_ui(
            msg="Select image directory (*.tif)",
            parent=self,
            path=os.path.dirname(csv_path),
        )

        if path is None:
            napari.utils.notifications.show_warning(
                f"No image directory selected. Exiting."
            )
            return

        filenames = []
        for fname in pd.unique(df["filename"].values):
            filenames += glob.glob(os.path.join(path, f"*{fname}*"))

        if len(filenames) == 0:
            napari.utils.notifications.show_warning(
                f"No images found in {path}. Exiting"
            )
            return

        self.load_data(df, filenames, path, csv_path)

    def load_data(
        self, df: pd.DataFrame, filenames: list[str], path: str, csv_path: str
    ):
        # set attributes
        self.filenames = filenames
        self.path = path
        self.csv_path = csv_path
        self.df = df

        # load filenames as single layer
        self.img_layer = self.viewer.open(filenames, stack=True)[0]

        # subscribe to dims events
        self.viewer.dims.events.current_step.connect(self._change_image)

        # load empty points layer
        self.pts_layer = self.viewer.add_points(
            name=CONFIGURATION["UI"]["POINT_LAYER_NAME"], 
            face_color=CONFIGURATION["UI"]["FACE_COLOR"], 
            size=CONFIGURATION["UI"]["POINT_SIZE"]
        )

        self.pts_layer.events.data.connect(self._features_updated)

        # update viewer
        self.update_viewer_to_image(0)

        # metadata
        self.img_layer.name = CONFIGURATION["UI"]["IMAGE_LAYER_NAME"]
        self.img_layer.metadata["path"] = path
        self.img_layer.metadata["csv_path"] = csv_path

        logging.info(f"Loaded {len(filenames)} images from {path}")
        logging.info(f"Loaded {len(df)} rows from {csv_path}")
        napari.utils.notifications.show_info(
            f"Loaded {len(filenames)} images from {path}"
        )

        self.update_ui_elements()

    def update_viewer_to_image(self, idx: int):
        self.viewer.dims.set_point(0, idx)

    def _change_image(self):
        """Change the currently selected image"""
        if self.sender() == self.spinBox_data_progress:
            idx = self.spinBox_data_progress.value()
            self.update_viewer_to_image(idx)
        else:
            idx = int(self.viewer.dims.point[0])
            self.spinBox_data_progress.blockSignals(True)
            self.spinBox_data_progress.setValue(idx)
            self.spinBox_data_progress.blockSignals(False)

        # clear features / points
        self.features = []
        self.feature = None
        self.pts_layer.data = []
        self._add_feature_mode()

        # update data from dataframe
        self.update_points_from_dataframe(idx)

    def _get_feature(self):
        """Get the currently selected feature"""
        self.feature = {"name": self.comboBox_features.currentText()}
        return self.feature

    def _add_feature_mode(self):
        """Add a feature to the currently selected image"""
        self.viewer.layers.selection.active = self.pts_layer
        self.pts_layer.mode = "add"

    def _edit_feature_mode(self):
        """Edit the currently selected feature"""
        self.viewer.layers.selection.active = self.pts_layer
        self.pts_layer.mode = "select"

    def _features_updated(self, event: Optional[Any] = None) -> None:
        if self.feature is None:
            self._get_feature()

        if self.feature is None:
            napari.utils.notifications.show_warning(f"No feature selected. Exiting.")
            return

        # get latest points
        points = self.pts_layer.data

        # get which point was moved
        index: list[int] = list(self.pts_layer.selected_data)

        def napari_pt_to_point(pt: np.ndarray, dtype=int) -> Point:
            return Point.from_list(np.flip(pt, axis=-1).astype(dtype))

        # add new feature
        if len(points) > len(self.features):
            # get latest feature
            feature = copy.deepcopy(self.feature)
            feature.update(napari_pt_to_point(points[-1]).to_dict())
            feature["color"] = get_feature_color(feature["name"])
            self.features.append(copy.deepcopy(feature))

        # feature was deleted
        elif len(points) < len(self.features):
            # deleted

            for idx in index[::-1]:
                logging.info(f"point deleted: {self.features[idx]['name']}")
                self.features.pop(idx)

        # feature was moved
        else:
            # moved
            for idx in index:
                self.features[idx].update(
                    copy.deepcopy(napari_pt_to_point(points[idx])).to_dict()
                )

        #
        text = {
            "string": [f["name"] for f in self.features],
            "color": CONFIGURATION["UI"]["TEXT_COLOR"],
            "translation": CONFIGURATION["UI"]["TEXT_TRANSLATION"],
        }

        self.pts_layer.text = text
        self.pts_layer.face_color = [f["color"] for f in self.features]

        # update dataframe, write to csv
        self.update_dataframe()

    def update_dataframe(self):
        """Update the dataframe with the latest features"""
        idx = int(self.viewer.dims.point[0])
        fname = self.filenames[idx]

        df: pd.DataFrame = self.df
        # get features
        features = self.features

        # if features is empty, add empty row to maintain image in dataframe
        if len(features) == 0:
            features = [{"name": np.nan, "x": np.nan, "y": np.nan}]

        # get dataframe for current image
        df_filt = df[df["filename"] == os.path.basename(fname)]

        # create dataframe
        df_new = pd.DataFrame(
            columns=[
                "filename",
                "feature",
                "px.x",
                "px.y",
                "pixelsize",
                "corrected",
                "method",
                "experiment",
            ]
        )
        df_new["filename"] = [os.path.basename(fname)] * len(features)
        df_new["feature"] = [f["name"] for f in features]
        df_new["px.x"] = [f["x"] for f in features]
        df_new["px.y"] = [f["y"] for f in features]
        df_new["pixelsize"] = [None] * len(features)  # pixel size is not known
        df_new["corrected"] = [True] * len(features)
        df_new["method"] = [df_filt["method"].values[0]] * len(features)
        df_new["experiment"] = [df_filt["experiment"].values[0]] * len(features)

        # update dataframe
        df.drop(df_filt.index, inplace=True)
        df = pd.concat([df, df_new], axis=0)
        df.reset_index(drop=True, inplace=True)  # mind your index

        # save dataframe
        df.to_csv(self.csv_path, index=False)

        # update dataframe
        self.df = df

    def update_points_from_dataframe(self, idx: int):
        """Update the points layer from the dataframe"""
        # get dataframe for current image
        fname = self.filenames[idx]
        df_filt = self.df[self.df["filename"] == os.path.basename(fname)]

        # initial empty df, then no features were added
        if len(df_filt) == 1 and pd.isna(df_filt["feature"].values[0]):
            napari.utils.notifications.show_info(
                f"No features yet for {os.path.basename(fname)}"
            )
            return

        # load features from dataframe
        for _, row in df_filt.iterrows():
            d = {
                "name": row["feature"],
                "x": row["px.x"],
                "y": row["px.y"],
                "color": get_feature_color(row["feature"]),
            }
            self.features.append(copy.deepcopy(d))

        # update points layer
        self.pts_layer.data = np.array([[f["y"], f["x"]] for f in self.features])

        # update text
        text = {
            "string": [f["name"] for f in self.features],
            "color": "white",
            "translation": np.array([-30, 0]),
        }

        self.pts_layer.text = text
        self.pts_layer.face_color = [f["color"] for f in self.features]
        napari.utils.notifications.show_info(
            f"Loaded {len(self.features)} features from {os.path.basename(fname)}"
        )


def main():
    viewer = napari.Viewer(ndisplay=2)
    fibsem_ui = FibsemFeatureLabellingUI(viewer=viewer)
    viewer.window.add_dock_widget(
        fibsem_ui,
        area="right",
        add_vertical_stretch=True,
        name=f"OpenFIBSEM v{fibsem.__version__} - Keypoint Labelling",
    )
    napari.run()


if __name__ == "__main__":
    main()



from fibsem import utils, calibration, acquire

from fibsem.structures import BeamType, FibsemManipulatorPosition
import matplotlib.pyplot as plt
import numpy as np
import sys

def main():
    microscope, settings = utils.setup_session()

    # if argv 1 is "debug", run in debug mode
    _DEBUG = False
    if len(sys.argv) > 1:
        if sys.argv[1] == "debug":
            _DEBUG = True

    if _DEBUG:
        microscope.set("scan_rotation", np.deg2rad(0), beam_type=BeamType.ION)
        microscope.move_manipulator_to_position_offset(offset=FibsemManipulatorPosition(), name="EUCENTRIC")
        microscope.move_manipulator_corrected(dx=-50e-6, dy=-150e-6, beam_type=BeamType.ELECTRON)

        settings.image.autocontrast = True
        settings.image.hfw = 900e-6
        eb_image, ib_image = acquire.take_reference_images(microscope, settings.image)

        fig, ax = plt.subplots(1, 2, figsize=(10, 5))
        ax[0].imshow(eb_image.data, cmap="gray")
        ax[1].imshow(ib_image.data, cmap="gray")

        plt.show()

    calibration._calibrate_manipulator_thermo(microscope, settings, None)

if __name__ == "__main__":
    main()


import sqlite3
from sqlite3 import Error

def create_table(conn, create_table_sql):
    """ create a table from the create_table_sql statement
    :param conn: Connection object
    :param create_table_sql: a CREATE TABLE statement
    :return:
    """
    try:
        c = conn.cursor()
        c.execute(create_table_sql)
    except Error as e:
        print(e)




def create_project(conn, project):
    """
    Create a new project into the projects table
    :param conn:
    :param project:
    :return: project id
    """
    sql = ''' INSERT INTO projects(name,date,user)
              VALUES(?,?,?) '''
    cur = conn.cursor()
    cur.execute(sql, project)
    conn.commit()
    return cur.lastrowid


def create_user(conn, user):
    """
    Create a new user into the users table
    :param conn:
    :param user:
    :return: user id
    """
    sql = ''' INSERT INTO users(name,email,password)
              VALUES(?,?,?) '''
    cur = conn.cursor()
    cur.execute(sql, user)
    conn.commit()
    return cur.lastrowid

def create_sample(conn, sample):
    """
    Create a new sample into the samples table
    :param conn:
    :param sample:
    :return: sample id
    """
    sql = ''' INSERT INTO samples(name,project_id,date,user)
              VALUES(?,?,?,?) '''
    cur = conn.cursor()
    cur.execute(sql, sample)
    conn.commit()
    return cur.lastrowid

def create_experiment(conn, experiment):
    """
    Create a new experiment into the experiments table
    :param conn:
    :param experiment:
    :return: experiment id
    """
    sql = ''' INSERT INTO experiments(name,project_id,date,user,sample_id,program,method,path)
              VALUES(?,?,?,?,?,?,?,?) '''
    cur = conn.cursor()
    cur.execute(sql, experiment)
    conn.commit()
    return cur.lastrowid

def create_connection(db_file):
    """ create a database connection to the SQLite database
        specified by the db_file
    :param db_file: database file
    :return: Connection object or None
    """
    conn = None
    try:
        conn = sqlite3.connect(db_file, 
        detect_types=sqlite3.PARSE_DECLTYPES |
                             sqlite3.PARSE_COLNAMES)
    except Error as e:
        print(e)

    return conn


#####
SQL_CREATE_PROJECTS_TABLE = """ CREATE TABLE IF NOT EXISTS projects (
                                    id INTEGER PRIMARY KEY,
                                    name VARCHAR(100) NOT NULL,
                                    date TIMESTAMP NOT NULL,
                                    user VARCHAR(100) NOT NULL
                                ); """

SQL_CREATE_USERS_TABLES = """CREATE TABLE IF NOT EXISTS users (
                                id INTEGER PRIMARY KEY,
                                name VARCHAR(100) NOT NULL,
                                email VARCHAR(100) NOT NULL,
                                password VARCHAR(100) NOT NULL
                            );"""


SQL_CREATE_SAMPLES_TABLE = """CREATE TABLE IF NOT EXISTS samples (
                                id INTEGER PRIMARY KEY,
                                name VARCHAR(100) NOT NULL,
                                project_id INTEGER NOT NULL,
                                date TIMESTAMP NOT NULL,
                                user VARCHAR(100) NOT NULL,
                                FOREIGN KEY (project_id) REFERENCES projects (id)
                                );"""


SQL_CREATE_EXPERIMENTS_TABLE = """CREATE TABLE IF NOT EXISTS experiments (
                                id INTEGER PRIMARY KEY,
                                name VARCHAR(100) NOT NULL,
                                project_id INTEGER NOT NULL,
                                date TIMESTAMP NOT NULL,
                                user VARCHAR(100) NOT NULL,
                                sample_id INTEGER NOT NULL,
                                program VARCHAR(100) NOT NULL,
                                method VARCHAR(100) NOT NULL,
                                path VARCHAR(100) NOT NULL,
                                FOREIGN KEY (project_id) REFERENCES projects (id)
                                FOREIGN KEY (sample_id) REFERENCES samples (id)
                            );"""



SQL_CREATE_HISTORY_TABLE = """CREATE TABLE IF NOT EXISTS history (
                                id INTEGER PRIMARY KEY,
                                petname VARCHAR(100) NOT NULL,
                                stage VARCHAR(100) NOT NULL,
                                start TIMESTAMP NOT NULL,
                                end TIMESTAMP NOT NULL,
                                duration FLOAT NOT NULL,
                                experiment_id INTEGER NOT NULL,
                                FOREIGN KEY (experiment_id) REFERENCES experiment (id)
                                
                            );"""

SQL_CREATE_STEPS_TABLE = """CREATE TABLE IF NOT EXISTS steps (
                                id INTEGER PRIMARY KEY,
                                petname VARCHAR(100) NOT NULL,
                                stage VARCHAR(100) NOT NULL,
                                step VARCHAR(100) NOT NULL,
                                step_n INTEGER NOT NULL,
                                timestamp TIMESTAMP NOT NULL,
                                duration FLOAT NOT NULL,
                                experiment_id INTEGER NOT NULL,
                                FOREIGN KEY (experiment_id) REFERENCES experiment (id)
                                
                            );"""



SQL_CREATE_DETECTIONS_TABLE = """CREATE TABLE IF NOT EXISTS detections (
                                id INTEGER PRIMARY KEY,
                                petname VARCHAR(100) NOT NULL,
                                stage VARCHAR(100) NOT NULL,
                                step VARCHAR(100) NOT NULL,
                                feature VARCHAR(100) NOT NULL,
                                px_x INTEGER NOT NULL,
                                px_y INTEGER NOT NULL,
                                dpx_x INTEGER NOT NULL,
                                dpx_y INTEGER NOT NULL,
                                dm_x FLOAT NOT NULL,
                                dm_y FLOAT NOT NULL,
                                is_correct BOOL NOT NULL,
                                beam_type VARCHAR(100) NOT NULL,
                                fname VARCHAR(100) NOT NULL,
                                timestamp TIMESTAMP NOT NULL,
                                experiment_id INTEGER NOT NULL,
                                FOREIGN KEY (experiment_id) REFERENCES experiment (id)
                                
                            );"""


SQL_CREATE_INTERACTIONS_TABLE = """CREATE TABLE IF NOT EXISTS interactions (
                                id INTEGER PRIMARY KEY,
                                petname VARCHAR(100) NOT NULL,
                                stage VARCHAR(100) NOT NULL,
                                step VARCHAR(100) NOT NULL,
                                type VARCHAR(100) NOT NULL,
                                subtype VARCHAR(100) NOT NULL, 
                                dm_x FLOAT NOT NULL,
                                dm_y FLOAT NOT NULL,
                                beam_type VARCHAR(100) NOT NULL,
                                timestamp TIMESTAMP NOT NULL,
                                experiment_id INTEGER NOT NULL,
                                FOREIGN KEY (experiment_id) REFERENCES experiment (id)
                                
                            );"""

from fibsem import config as cfg
import os

def _create_database():

    # create / connect to db
    database = cfg.DATABASE_PATH
    conn = create_connection(database)

    # create tables
    create_table(conn, SQL_CREATE_PROJECTS_TABLE)
    create_table(conn, SQL_CREATE_USERS_TABLES)
    create_table(conn, SQL_CREATE_SAMPLES_TABLE)
    create_table(conn, SQL_CREATE_EXPERIMENTS_TABLE)
    create_table(conn, SQL_CREATE_HISTORY_TABLE)
    create_table(conn, SQL_CREATE_STEPS_TABLE)
    create_table(conn, SQL_CREATE_DETECTIONS_TABLE)
    create_table(conn, SQL_CREATE_INTERACTIONS_TABLE)
    print("Database created successfully")
    conn.commit()
    conn.close()



import json
from abc import ABC, abstractmethod
from dataclasses import dataclass

import yaml
import numpy as np
from fibsem import constants 
from fibsem.structures import (Point, 
    FibsemPatternSettings, FibsemRectangleSettings, 
    FibsemCircleSettings, FibsemBitmapSettings,
    FibsemLineSettings, CrossSectionPattern)


def check_keys(protocol: dict, required_keys: list[str]) -> bool:
    return all([k in protocol.keys() for k in required_keys])

# TODO: define the configuration for each key,
# e.g. 
# "width": {"type": "float", "min": 0, "max": 1000, "default": 100, "description": "Width of the rectangle"}
# "cross_section": {"type": "str", "options": [cs.name for cs in CrossSectionPattern], "default": "Rectangle", "description": "Cross section of the milling pattern"}

    
REQUIRED_KEYS = {
    "Rectangle": ("width", "height", "depth", "rotation","passes", "cleaning_cross_section","scan_direction", "cross_section"),
    "Line": ("start_x", "end_x", "start_y", "end_y", "depth"),
    "Circle": ("radius", "depth","cleaning_cross_section"),
    "Trench": (
        "lamella_width",
        "lamella_height",
        "trench_height",
        "size_ratio",
        "offset",
        "depth",
        "cleaning_cross_section",
        "cross_section",
    ),
    "Horseshoe": (
        "lamella_width",
        "lamella_height",
        "trench_height",
        "size_ratio",
        "offset",
        "side_offset",
        "side_width",
        "depth",
        "scan_direction",
        "cross_section",
    ),
    "HorseshoeVertical": (
        "width",
        "height",
        "side_trench_width",
        "top_trench_height",
        "depth",
        "scan_direction",
        "inverted",
        "cross_section", 
    ),
    "Fiducial": ("height", "width", "depth", "rotation","cleaning_cross_section", "cross_section"),
    "Undercut": (
        "height",
        "width",
        "depth",
        "trench_width",
        "rhs_height",
        "h_offset",
        "cleaning_cross_section",
        "cross_section",
    ),
    "MicroExpansion": (
        "height",
        "width",
        "depth",
        "distance",
    ),
    "ArrayPattern": ("height", "width", "depth", "n_columns", "n_rows", 
                    "pitch_vertical", "pitch_horizontal", 
                    "passes", "scan_direction", "cross_section"),
    "WaffleNotch": (
        "vheight",
        "vwidth",
        "hheight",
        "hwidth",
        "depth",
        "distance",
        "cross_section",
    ),
    "Clover": ("radius", "depth"),
    "TriForce": ("height", "width", "depth"),
    "Trapezoid": ("inner_width", "outer_width", "trench_height", "depth", "distance", "n_rectangles", "overlap"),
}


####### Combo Patterns

@dataclass
class BasePattern(ABC):
    name: str = "BasePattern"
    required_keys: tuple[str] = ()
    patterns = None
    protocol = None
    point: Point = Point()

    @abstractmethod
    def define(
        self, protocol: dict, point: Point = Point()
    ) -> list[FibsemPatternSettings]:
        pass

    def to_dict(self):
        return {
            "name": self.name,
            "required_keys": self.required_keys,
            "protocol": self.protocol,
            "point": self.point.to_dict() if self.point is not None else None,
        }

    @classmethod
    def from_dict(cls, data):
        pattern = cls(
            name=data["name"],
            required_keys=REQUIRED_KEYS[data["name"]],
        )
        pattern.protocol = data["protocol"]
        pattern.point = (
            Point.from_dict(data["point"]) if data["point"] is not None else Point()
        )
        return pattern


@dataclass
class BitmapPattern(BasePattern):
    name: str = "BitmapPattern"
    required_keys: tuple[str] = ("width", "height", "depth", "rotation","path")
    patterns = None
    protocol = None

    def define(
            self, protocol: dict, point: Point = Point()
    ) -> list[FibsemPatternSettings]:
        protocol["centre_x"] = point.x
        protocol["centre_y"] = point.y
        protocol["pattern"] = "BitmapPattern"  
        protocol["rotation"] = protocol.get("rotation", 0)
        protocol["cleaning_cross_section"] = protocol.get("cleaning_cross_section", False)
        protocol["scan_direction"] = protocol.get("scan_direction", "TopToBottom")
        protocol["path"] = protocol.get("path", 'bmp_path')
        self.patterns = [FibsemBitmapSettings.from_dict(protocol)]
        self.protocol = protocol
        return self.patterns


@dataclass
class RectanglePattern(BasePattern):
    name: str = "Rectangle"
    required_keys: tuple[str] = REQUIRED_KEYS["Rectangle"]
    patterns = None
    protocol = None
    point = None

    def define(
        self, protocol: dict, point: Point = Point()
    ) -> list[FibsemRectangleSettings]:
        protocol["centre_x"] = point.x
        protocol["centre_y"] = point.y
        protocol["pattern"] = "Rectangle"  # redundant now
        protocol["rotation"] = protocol.get("rotation", 0) * constants.DEGREES_TO_RADIANS
        protocol["cleaning_cross_section"] = protocol.get(
            "cleaning_cross_section", False
        )
        protocol["cross_section"] = protocol.get("cross_section", "Rectangle")
        protocol["scan_direction"] = protocol.get("scan_direction", "TopToBottom")
        protocol["passes"] = protocol.get("passes", None)
        self.patterns = [FibsemRectangleSettings.from_dict(protocol)]
        self.protocol = protocol
        self.point = point
        return self.patterns


@dataclass
class LinePattern(BasePattern):
    name: str = "Line"
    required_keys: tuple[str] = REQUIRED_KEYS["Line"]
    patterns = None
    protocol = None
    point = None

    def define(
        self, protocol: dict, point: Point = Point()
    ) -> list[FibsemLineSettings]:
        protocol["pattern"] = "Line"  # redundant now
        protocol["centre_x"] = point.x
        protocol["centre_y"] = point.y
        protocol["cleaning_cross_section"] = protocol.get(
            "cleaning_cross_section", False
        )
        protocol["scan_direction"] = protocol.get("scan_direction", "TopToBottom")
        self.patterns = [FibsemLineSettings.from_dict(protocol)]
        self.protocol = protocol
        self.point = point
        return self.patterns


@dataclass
class CirclePattern(BasePattern):
    name: str = "Circle"
    required_keys: tuple[str] = REQUIRED_KEYS["Circle"]
    patterns = None
    protocol = None
    point = None

    def define(
        self, protocol: dict, point: Point = Point()
    ) -> list[FibsemCircleSettings]:
        protocol["centre_x"] = point.x
        protocol["centre_y"] = point.y
        protocol["pattern"] = "Circle"  # redundant now
        protocol["start_angle"] = 0
        protocol["end_angle"] = 360
        protocol["rotation"] = 0
        protocol["cleaning_cross_section"] = protocol.get(
            "cleaning_cross_section", False
        )
        protocol["scan_direction"] = protocol.get("scan_direction", "TopToBottom")
        self.patterns = [FibsemCircleSettings.from_dict(protocol)]
        self.protocol = protocol
        self.point = point
        return self.patterns

@dataclass
class AnnulusPattern(BasePattern):
    name: str = "Annulus"
    required_keys: tuple[str] = ("thickness", "radius", "depth")
    patterns = None
    protocol = None
    point = None
    def define(
            self, protocol: dict, point: Point = Point()
    ) -> list[FibsemCircleSettings]:
        
        protocol["centre_x"] = point.x
        protocol["centre_y"] = point.y
        protocol["pattern"] = "Annulus"  # redundant now
        protocol["start_angle"] = 0
        protocol["end_angle"] = 360
        protocol["rotation"] = 0
        protocol["cleaning_cross_section"] = protocol.get("cleaning_cross_section", False)
        protocol["scan_direction"] = protocol.get("scan_direction", "TopToBottom")


        self.patterns = [FibsemCircleSettings.from_dict(protocol)]
        self.protocol = protocol
        self.point = point
        return self.patterns


@dataclass
class TrenchPattern(BasePattern):
    name: str = "Trench"
    required_keys: tuple[str] = REQUIRED_KEYS["Trench"]
    patterns = None
    protocol = None
    point = None

    def define(
        self, protocol: dict, point: Point = Point()
    ) -> list[FibsemRectangleSettings]:
        check_keys(protocol, self.required_keys)

        lamella_width = protocol["lamella_width"]
        lamella_height = protocol["lamella_height"]
        trench_height = protocol["trench_height"]
        upper_trench_height = trench_height / max(protocol["size_ratio"], 1.0)
        lower_trench_height = trench_height * min(protocol["size_ratio"], 1.0)
        offset = protocol["offset"]
        depth = protocol["depth"]
        use_cleaning_cross_section = protocol.get("cleaning_cross_section", True)
        cross_section = CrossSectionPattern[protocol.get("cross_section", "Rectangle")]

        centre_upper_y = point.y + (
            lamella_height / 2 + upper_trench_height / 2 + offset
        )
        centre_lower_y = point.y - (lamella_height / 2 + lower_trench_height / 2 + offset)

        # mill settings
        lower_pattern_settings = FibsemRectangleSettings(
            width=lamella_width,
            height=lower_trench_height,
            depth=depth,
            centre_x=point.x,
            centre_y=centre_lower_y,
            cleaning_cross_section=use_cleaning_cross_section,
            scan_direction="BottomToTop",
            cross_section = cross_section

        )

        upper_pattern_settings = FibsemRectangleSettings(
            width=lamella_width,
            height=upper_trench_height,
            depth=depth,
            centre_x=point.x,
            centre_y=centre_upper_y,
            cleaning_cross_section=use_cleaning_cross_section,
            scan_direction="TopToBottom",
            cross_section = cross_section
        )

        self.patterns = [lower_pattern_settings, upper_pattern_settings]
        self.protocol = protocol
        self.point = point
        return self.patterns


@dataclass
class HorseshoePattern(BasePattern):
    name: str = "Horseshoe"
    required_keys: tuple[str] = REQUIRED_KEYS["Horseshoe"]
    patterns = None
    # ref: "horseshoe" terminology https://www.researchgate.net/publication/351737991_A_Modular_Platform_for_Streamlining_Automated_Cryo-FIB_Workflows#pf14
    protocol = None
    point = None

    def define(
        self, protocol: dict, point: Point = Point()
    ) -> list[FibsemRectangleSettings]:
        """Calculate the trench milling patterns"""

        check_keys(protocol, self.required_keys)

        lamella_width = protocol["lamella_width"]
        lamella_height = protocol["lamella_height"]
        trench_height = protocol["trench_height"]
        upper_trench_height = trench_height / max(protocol["size_ratio"], 1.0)
        offset = protocol["offset"]
        depth = protocol["depth"]
        cross_section = CrossSectionPattern[protocol.get("cross_section", "Rectangle")]


        centre_upper_y = point.y + (
            lamella_height / 2 + upper_trench_height / 2 + offset
        )
        centre_lower_y = point.y - (lamella_height / 2 + trench_height / 2 + offset)

        lower_pattern = FibsemRectangleSettings(
            width=lamella_width,
            height=trench_height,
            depth=depth,
            centre_x=point.x,
            centre_y=centre_lower_y,
            cleaning_cross_section=False,
            scan_direction="BottomToTop",
            cross_section = cross_section
        )

        upper_pattern = FibsemRectangleSettings(
            width=lamella_width,
            height=upper_trench_height,
            depth=depth,
            centre_x=point.x,
            centre_y=centre_upper_y,
            cleaning_cross_section=False,
            scan_direction="TopToBottom",
            cross_section = cross_section
        )

        side_pattern = FibsemRectangleSettings(
            width=protocol["side_width"],
            height=lamella_height + offset,
            depth=depth,
            centre_x=point.x
            - protocol["side_offset"]
            + (lamella_width / 2 - protocol["side_width"] / 2),
            centre_y=point.y,
            cleaning_cross_section=False,
            scan_direction="TopToBottom",
            cross_section=cross_section
        )

        self.patterns = [lower_pattern, upper_pattern, side_pattern]
        self.protocol = protocol
        self.point = point
        return self.patterns



@dataclass
class HorseshoePatternVertical(BasePattern):
    name: str = "HorseshoeVertical"
    required_keys: tuple[str] = REQUIRED_KEYS["HorseshoeVertical"]
    patterns = None
    # ref: "horseshoe" terminology https://www.researchgate.net/publication/351737991_A_Modular_Platform_for_Streamlining_Automated_Cryo-FIB_Workflows#pf14
    protocol = None
    point = None

    def define(
        self, protocol: dict, point: Point = Point()
    ) -> list[FibsemRectangleSettings]:
        """Calculate the horseshoe vertical milling patterns"""

        check_keys(protocol, self.required_keys)

        width = protocol["width"]
        height = protocol["height"]
        trench_width = protocol["side_trench_width"]
        upper_trench_height = protocol["top_trench_height"]
        depth = protocol["depth"]
        scan_direction = protocol.get("scan_direction", "TopToBottom")
        inverted = protocol.get("inverted", False)
        cross_section=CrossSectionPattern[protocol.get("cross_section", "Rectangle")]

        left_pattern = FibsemRectangleSettings(
            width=trench_width,
            height=height,
            depth=depth,
            centre_x=point.x - (width / 2) - (trench_width / 2),
            centre_y=point.y,
            cleaning_cross_section=False,
            scan_direction=scan_direction,
            cross_section=cross_section
        )

        right_pattern = FibsemRectangleSettings(
            width=trench_width,
            height=height,
            depth=depth,
            centre_x=point.x + (width / 2) + (trench_width / 2),
            centre_y=point.y,
            cleaning_cross_section=False,
            scan_direction=scan_direction,
            cross_section=cross_section
        )
        y_offset = (height / 2) + (upper_trench_height / 2)
        if inverted:
            y_offset = -y_offset
        upper_pattern = FibsemRectangleSettings(
            width=width + (2 * trench_width),
            height=upper_trench_height,
            depth=depth,
            centre_x=point.x,
            centre_y=point.y + y_offset,
            cleaning_cross_section=False,
            scan_direction=scan_direction,
            cross_section=cross_section
        )
        
        self.patterns = [left_pattern,right_pattern, upper_pattern]
        self.protocol = protocol
        self.point = point
        return self.patterns


@dataclass
class FiducialPattern(BasePattern):
    name: str = "Fiducial"
    required_keys: tuple[str] = REQUIRED_KEYS["Fiducial"]
    patterns = None
    protocol = None
    point = None

    def define(
        self, protocol: dict, point: Point = Point()
    ) -> list[FibsemRectangleSettings]:
        import numpy as np

        # ?
        protocol["centre_x"] = point.x
        protocol["centre_y"] = point.y
        protocol["pattern"] = "Rectangle"
        protocol["cleaning_cross_section"] = protocol.get(
            "cleaning_cross_section", False
        )
        protocol["cross_section"] = protocol.get("cross_section", "Rectangle")
        protocol["scan_direction"] = protocol.get("scan_direction", "TopToBottom")

        left_pattern = FibsemRectangleSettings.from_dict(protocol)
        from fibsem import constants 
        left_pattern.rotation = protocol["rotation"] * constants.DEGREES_TO_RADIANS
        right_pattern = FibsemRectangleSettings.from_dict(protocol)
        right_pattern.rotation = left_pattern.rotation + np.deg2rad(90)

        self.patterns = [left_pattern, right_pattern]
        self.protocol = protocol
        self.point = point
        return self.patterns


@dataclass
class UndercutPattern(BasePattern):
    name: str = "Undercut"
    required_keys: tuple[str] = REQUIRED_KEYS["Undercut"]
    patterns = None
    protocol = None
    point = None

    def define(
        self, protocol: dict, point: Point = Point()
    ) -> list[FibsemRectangleSettings]:
        check_keys(protocol, self.required_keys)

        jcut_rhs_height = protocol["rhs_height"]
        jcut_lamella_height = protocol["height"]
        jcut_width = protocol["width"]
        jcut_trench_thickness = protocol["trench_width"]
        jcut_depth = protocol["depth"]
        jcut_h_offset = protocol["h_offset"]
        cross_section = CrossSectionPattern[protocol.get("cross_section", "Rectangle")]
        use_cleaning_cross_section = protocol.get("cleaning_cross_section", False)

        # top_jcut
        jcut_top_centre_x = point.x + jcut_width / 2 - jcut_h_offset
        jcut_top_centre_y = point.y + jcut_lamella_height
        jcut_top_width = jcut_width
        jcut_top_height = jcut_trench_thickness
        jcut_top_depth = jcut_depth

        top_pattern = FibsemRectangleSettings(
            width=jcut_top_width,
            height=jcut_top_height,
            depth=jcut_top_depth,
            centre_x=point.x,
            centre_y=point.y,
            cleaning_cross_section=use_cleaning_cross_section,
            scan_direction="TopToBottom",
            cross_section=cross_section
        )
        # rhs jcut
        jcut_rhs_centre_x = point.x + (jcut_width / 2) - jcut_trench_thickness / 2
        jcut_rhs_centre_y = point.y - (jcut_rhs_height / 2) + jcut_trench_thickness / 2
        jcut_rhs_width = jcut_trench_thickness
        jcut_rhs_height = jcut_rhs_height
        jcut_rhs_depth = jcut_depth

        rhs_pattern = FibsemRectangleSettings(
            width=jcut_rhs_width,
            height=jcut_rhs_height,
            depth=jcut_rhs_depth,
            centre_x=jcut_rhs_centre_x,
            centre_y=jcut_rhs_centre_y,
            cleaning_cross_section=use_cleaning_cross_section,
            scan_direction="TopToBottom",
            cross_section = CrossSectionPattern[protocol.get("cross_section", "Rectangle")]
        )



        self.patterns = [top_pattern, rhs_pattern]
        self.protocol = protocol
        self.point = point
        return self.patterns


@dataclass
class MicroExpansionPattern(BasePattern):
    name: str = "MicroExpansion"
    required_keys: tuple[str] = REQUIRED_KEYS["MicroExpansion"]
    patterns = None
    protocol = None
    point = None

    # ref: https://www.nature.com/articles/s41467-022-29501-3
    def define(
        self, protocol: dict, point: Point = Point()
    ) -> list[FibsemRectangleSettings]:
        """
        Draw the microexpansion joints for stress relief of lamella.

        Args:
            microscope (FibsemMicroscope): OpenFIBSEM microscope instance
            protocol (dict): Contains a dictionary of the necessary values for drawing the joints.
            protocol (dict): Lamella protocol

        Returns:
            patterns: list[FibsemPatternSettings]
        """
        check_keys(protocol, self.required_keys)

        width = protocol["width"]
        height = protocol["height"]
        depth = protocol["depth"]  # lamella milling depth

        left_pattern_settings = FibsemRectangleSettings(
            width=width,
            height=height,
            depth=depth,
            centre_x=point.x -  protocol["distance"] ,
            centre_y=point.y,
            cleaning_cross_section=True,
            scan_direction="TopToBottom",
        )

        right_pattern_settings = FibsemRectangleSettings(
            width=width,
            height=height,
            depth=depth,
            centre_x=point.x + protocol["distance"],
            centre_y=point.y,
            cleaning_cross_section=True,
            scan_direction="TopToBottom",
        )

        self.patterns = [left_pattern_settings, right_pattern_settings]
        self.protocol = protocol
        self.point = point
        return self.patterns

@dataclass
class ArrayPattern(BasePattern):
    name: str = "ArrayPattern"
    required_keys: tuple[str] = REQUIRED_KEYS["ArrayPattern"]
    patterns = None
    # ref: spotweld terminology https://www.researchgate.net/publication/351737991_A_Modular_Platform_for_Streamlining_Automated_Cryo-FIB_Workflows#pf14
    # ref: weld cross-section/ passes: https://www.nature.com/articles/s41592-023-02113-5
    protocol = None
    point = None

    def define(
        self, protocol: dict, point: Point = Point()
    ) -> list[FibsemRectangleSettings]:
        check_keys(protocol, self.required_keys)

        width = protocol["width"]
        height = protocol["height"]
        depth = protocol["depth"]
        n_columns = int(protocol["n_columns"])
        n_rows = int(protocol["n_rows"])
        pitch_vertical = protocol["pitch_vertical"]
        pitch_horizontal = protocol["pitch_horizontal"]
        rotation = protocol.get("rotation", 0)
        passes = protocol.get("passes", 1)
        scan_direction = protocol.get("scan_direction", "TopToBottom")
        passes = int(passes) if passes is not None else None
        cross_section = CrossSectionPattern[protocol.get("cross_section", "Rectangle")]

        # create a 2D array of points
        points = []
        for i in range(n_columns):
            for j in range(n_rows):
                points.append(
                    Point(
                        point.x + (i - (n_columns - 1) / 2) * pitch_horizontal,
                        point.y + (j - (n_rows - 1) / 2) * pitch_vertical,
                    )
                )
        # create patterns
        patterns = []
        for point in points:
            pattern_settings = FibsemRectangleSettings(
                width=width,
                height=height,
                depth=depth,
                centre_x=point.x,
                centre_y=point.y,  
                cleaning_cross_section=False,
                scan_direction=scan_direction,

                rotation=rotation,
                passes=passes,
                cross_section=cross_section,
            )
            patterns.append(pattern_settings)

        self.patterns = patterns
        self.protocol = protocol
        self.point = point
        return self.patterns




@dataclass
class WaffleNotchPattern(BasePattern):
    name: str = "WaffleNotch"
    required_keys: tuple[str] = REQUIRED_KEYS["WaffleNotch"]
    patterns = None
    protocol = None
    point = None
    # ref: https://www.nature.com/articles/s41467-022-29501-3

    def define(
        self, protocol: dict, point: Point = Point()
    ) -> list[FibsemRectangleSettings]:
        check_keys(protocol, self.required_keys)

        vwidth = protocol["vwidth"]
        vheight = protocol["vheight"]
        hwidth = protocol["hwidth"]
        hheight = protocol["hheight"]
        depth = protocol["depth"]
        distance = protocol["distance"]
        cross_section = CrossSectionPattern[protocol.get("cross_section", "Rectangle")]

        # five patterns
        top_vertical_pattern = FibsemRectangleSettings(
            width=vwidth,
            height=vheight,
            depth=depth,
            centre_x=point.x,
            centre_y=point.y - distance / 2 - vheight / 2 + hheight / 2,
            cleaning_cross_section=False,
            scan_direction="TopToBottom",
            cross_section=cross_section
        )

        bottom_vertical_pattern = FibsemRectangleSettings(
            width=vwidth,
            height=vheight,
            depth=depth,
            centre_x=point.x,
            centre_y=point.y + distance / 2 + vheight / 2 - hheight / 2,
            cleaning_cross_section=False,
            scan_direction="BottomToTop",
            cross_section=cross_section
        )

        top_horizontal_pattern = FibsemRectangleSettings(
            width=hwidth,
            height=hheight,
            depth=depth,
            centre_x=point.x + hwidth / 2 + vwidth / 2,
            centre_y=point.y - distance / 2,
            cleaning_cross_section=False,
            scan_direction="TopToBottom",
            cross_section=cross_section
        )

        bottom_horizontal_pattern = FibsemRectangleSettings(
            width=hwidth,
            height=hheight,
            depth=depth,
            centre_x=point.x + hwidth / 2 + vwidth / 2,
            centre_y=point.y + distance / 2,
            cleaning_cross_section=False,
            scan_direction="BottomToTop",
            cross_section=cross_section
        )

        centre_vertical_pattern = FibsemRectangleSettings(
            width=vwidth,
            height=distance + hheight,
            depth=depth,
            centre_x=point.x + hwidth + vwidth,
            centre_y=point.y,
            cleaning_cross_section=False,
            scan_direction="TopToBottom",
            cross_section=cross_section
        )

        self.patterns = [
            top_vertical_pattern,
            bottom_vertical_pattern,
            top_horizontal_pattern,
            bottom_horizontal_pattern,
            centre_vertical_pattern,
        ]
        self.protocol = protocol
        self.point = point
        return self.patterns


@dataclass
class CloverPattern(BasePattern):
    name: str = "Clover"
    required_keys: tuple[str] = REQUIRED_KEYS["Clover"]
    patterns = None
    protocol = None
    point = None

    def define(
        self, protocol: dict, point: Point = Point()
    ) -> list[FibsemPatternSettings]:
        check_keys(protocol, self.required_keys)

        radius = protocol["radius"]
        depth = protocol["depth"]

        # three leaf clover pattern

        top_pattern = FibsemCircleSettings(
            radius=radius,
            depth=depth,
            centre_x=point.x,
            centre_y=point.y + radius,
        )

        right_pattern = FibsemCircleSettings(
            radius=radius,
            depth=depth,
            centre_x=point.x + radius,
            centre_y=point.y,
        )

        left_pattern = FibsemCircleSettings(
            radius=radius,
            depth=depth,
            centre_x=point.x - radius,
            centre_y=point.y,
        )

        stem_pattern = FibsemRectangleSettings(
            width=radius / 4,
            height=radius * 2,
            depth=depth,
            centre_x=point.x,
            centre_y=point.y - radius,
            cleaning_cross_section=False,
            scan_direction="TopToBottom",
        )

        self.patterns = [top_pattern, right_pattern, left_pattern, stem_pattern]
        self.protocol = protocol
        self.point = point
        return self.patterns


@dataclass
class TriForcePattern(BasePattern):
    name: str = "TriForce"
    required_keys: tuple[str] = REQUIRED_KEYS["TriForce"]
    patterns = None
    protocol = None
    point = None

    def define(
        self, protocol: dict, point: Point = Point()
    ) -> list[FibsemPatternSettings]:
        check_keys(protocol, self.required_keys)

        height = protocol["height"]
        width = protocol["width"]
        depth = protocol["depth"]

        # triforce pattern
        angle = 30

        self.patterns = []

        # centre of each triangle
        points = [
            Point(point.x, point.y + height),
            Point(point.x - height / 2, point.y),
            Point(point.x + height / 2, point.y),
        ]

        for point in points:

            left_pattern, right_pattern, bottom_pattern = create_triangle_patterns(width=width, height=height, depth=depth, point=point, angle=angle)

            self.patterns.append(left_pattern)
            self.patterns.append(right_pattern)
            self.patterns.append(bottom_pattern)
        self.protocol = protocol
        self.point = point
        return self.patterns


@dataclass
class TrapezoidPattern(BasePattern):
    name: str = "Trapezoid"
    required_keys: tuple[str] = REQUIRED_KEYS["Trapezoid"]
    patterns = None
    protocol = None
    point = None

    def define(self, protocol: dict, point: Point = Point()) -> list[FibsemPatternSettings]:
        check_keys(protocol, self.required_keys)
        self.patterns = []
        width_increments = (protocol["outer_width"] - protocol["inner_width"]) / (protocol["n_rectangles"]-1)
        dict = {}
        dict["trench_height"] = protocol["trench_height"] / protocol["n_rectangles"] * (1+protocol["overlap"])
        dict["depth"] = protocol["depth"]
        # bottom half
        for i in range(int(protocol["n_rectangles"])):
            dict["width"] = protocol["outer_width"] - i * width_increments
            pattern = RectanglePattern()
            y = point.y + (i * dict["trench_height"] * (1-protocol["overlap"])) - protocol["distance"] - protocol["trench_height"]
            centre = Point(point.x, y)
            pattern = FibsemRectangleSettings(
                width=dict["width"],
                height=dict["trench_height"],
                depth=dict["depth"],
                centre_x=centre.x,
                centre_y=centre.y,
                cleaning_cross_section=False,
                scan_direction="BottomToTop",
            )
            self.patterns.append(deepcopy(pattern))
        # top half
        for i in range(int(protocol["n_rectangles"])):
            dict["width"] = protocol["outer_width"] - i * width_increments
            pattern = RectanglePattern()
            y = point.y - (i * dict["trench_height"] * (1-protocol["overlap"])) + protocol["distance"] + protocol["trench_height"]
            centre = Point(point.x, y)
            pattern = FibsemRectangleSettings(
                width=dict["width"],
                height=dict["trench_height"],
                depth=dict["depth"],
                centre_x=centre.x,
                centre_y=centre.y,
                cleaning_cross_section=False,
                scan_direction="TopToBottom",
            )
            self.patterns.append(deepcopy(pattern))
        self.protocol = protocol
        self.point = point
        return self.patterns


__PATTERNS__ = [
    RectanglePattern,
    LinePattern,
    CirclePattern,
    TrenchPattern,
    HorseshoePattern,
    HorseshoePatternVertical,
    UndercutPattern,
    FiducialPattern,
    ArrayPattern,
    MicroExpansionPattern,
    WaffleNotchPattern,
    CloverPattern,
    TriForcePattern,
    BitmapPattern,
    AnnulusPattern,
    TrapezoidPattern,
]


def get_pattern(name: str) -> BasePattern:
    for pattern in __PATTERNS__:
        if pattern.name == name:
            return pattern()

    raise ValueError(f"Pattern {name} not found.")


def get_pattern_names() -> list[str]:
    return [pattern.name for pattern in __PATTERNS__]


def get_pattern_required_keys(name: str) -> tuple[str]:
    pattern = get_pattern(name)
    return pattern.required_keys


def get_pattern_required_keys_dict() -> dict[str, list[str]]:
    return {pattern.name: pattern.required_keys for pattern in __PATTERNS__}


def get_pattern_required_keys_json() -> str:
    return json.dumps(get_pattern_required_keys_dict())


def get_pattern_required_keys_yaml() -> str:
    return yaml.dump(get_pattern_required_keys_dict())


def create_triangle_patterns(
    width: float, height: float, depth: float, angle: float = 30, point: Point = Point()
) -> list[FibsemRectangleSettings]:
    h_offset = height / 2 * np.sin(np.deg2rad(angle))

    left_pattern = FibsemRectangleSettings(
        width=width,
        height=height,
        depth=depth,
        rotation=np.deg2rad(-angle),
        centre_x=point.x - h_offset,
        centre_y=point.y,
        cleaning_cross_section=False,
        scan_direction="LeftToRight",
    )

    right_pattern = FibsemRectangleSettings(
        width=width,
        height=height,
        depth=depth,
        rotation=np.deg2rad(angle),
        centre_x=point.x + h_offset,
        centre_y=point.y,
        cleaning_cross_section=False,
        scan_direction="RightToLeft",
    )

    bottom_pattern = FibsemRectangleSettings(
        width=width,
        height=height,
        depth=depth,
        rotation=np.deg2rad(90),
        centre_x=point.x,
        centre_y=point.y - height / 2,
        cleaning_cross_section=False,
        scan_direction="BottomToTop",
    )

    return [left_pattern, right_pattern, bottom_pattern]


from fibsem.utils import FibsemMillingSettings

# TODO: move to structures
@dataclass
class FibsemMillingStage:
    name: str = "Milling Stage"
    num: int = 0
    milling: FibsemMillingSettings = FibsemMillingSettings()
    pattern: BasePattern = get_pattern("Rectangle")

    def to_dict(self):
        return {
            "name": self.name,
            "num": self.num,
            "milling": self.milling.to_dict(),
            "pattern": self.pattern.to_dict(),
        }

    @classmethod
    def from_dict(cls, data: dict):
        return cls(
            name=data["name"],
            num=data["num"],
            milling=FibsemMillingSettings.from_dict(data["milling"]),
            pattern=get_pattern(data["pattern"]["name"]).from_dict(data["pattern"]),
        )


PROTOCOL_MILL_MAP = {
    "cut": RectanglePattern,
    "fiducial": FiducialPattern,
    "flatten": RectanglePattern,
    "undercut": UndercutPattern,
    "horseshoe": HorseshoePattern,
    "lamella": TrenchPattern,
    "polish_lamella": TrenchPattern,
    "thin_lamella": TrenchPattern,
    "sever": RectanglePattern,
    "sharpen": RectanglePattern,
    "needle": RectanglePattern,
    "copper_release": HorseshoePattern,
    "serial_trench": HorseshoePattern,
    "serial_undercut": RectanglePattern,
    "serial_sever": RectanglePattern,
    "lamella_sever": RectanglePattern,
    "lamella_polish": TrenchPattern,
    "trench": TrenchPattern,
    "notch": WaffleNotchPattern,
    "microexpansion": MicroExpansionPattern,
    "clover": CloverPattern,
    "autolamella": TrenchPattern,
    "MillUndercut": RectanglePattern,
    "rectangle": RectanglePattern,
    "MillRoughCut": TrenchPattern,
    "MillRegularCut": TrenchPattern,
    "MillPolishingCut": TrenchPattern,

}


def _get_pattern(key: str, protocol: dict, point: Point = Point()) -> BasePattern:
    if "type" in protocol:
        pattern = get_pattern(protocol["type"])
    else:
        pattern = PROTOCOL_MILL_MAP[key]()
    pattern.define(protocol, point=point)
    return pattern


def _get_stage(key, protocol: dict, point: Point = Point(), i: int = 0) -> FibsemMillingStage:
    pattern = _get_pattern(key, protocol, point=point)
    mill_settings = FibsemMillingSettings(
        milling_current=protocol["milling_current"], 
        hfw=float(protocol["hfw"]),
        application_file=protocol.get("application_file", "Si"),
        preset=protocol.get("preset", "30 keV; 20 nA"))

    # milling stage name
    name = protocol.get("name", f"{key.title()} {i+1:02d}")

    stage = FibsemMillingStage(
        name=name, num=i, milling=mill_settings, pattern=pattern
    )
    return stage

from typing import Optional, Union

def get_milling_stages(key, protocol, point: Union[Point, list[Point]] = Point()):
    
    # TODO: maybe add support for defining point per stages?

    # convert point to list of points, same length as stages

    if "stages" in protocol[key]:
        stages = []
        for i, pstage in enumerate(protocol[key]["stages"]):
            
            if isinstance(point, list):
                pt = point[i]
            else:
                pt = point

            stages.append(_get_stage(key, pstage, point=pt, i=i))
    else:
        stages = [_get_stage(key, protocol[key], point=point)]
    return stages


from copy import deepcopy
def get_protocol_from_stages(stages: list):
    protocol = {}  
    protocol["stages"] = []

    if not isinstance(stages, list):
        stages = [stages]

    for stage in stages:
        # join two dicts union
        ddict = {**stage.to_dict()["milling"], 
            **stage.to_dict()["pattern"]["protocol"], 
            "type": stage.to_dict()["pattern"]["name"]}
        protocol["stages"].append(deepcopy(ddict))

    return protocol


# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'FibsemManipulatorWidget.ui'
#
# Created by: PyQt5 UI code generator 5.15.9
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_Form(object):
    def setupUi(self, Form):
        Form.setObjectName("Form")
        Form.resize(350, 489)
        self.gridLayout_2 = QtWidgets.QGridLayout(Form)
        self.gridLayout_2.setObjectName("gridLayout_2")
        self.dZ_spinbox = QtWidgets.QDoubleSpinBox(Form)
        self.dZ_spinbox.setMinimum(-1000.0)
        self.dZ_spinbox.setMaximum(1000.0)
        self.dZ_spinbox.setObjectName("dZ_spinbox")
        self.gridLayout_2.addWidget(self.dZ_spinbox, 8, 1, 1, 1)
        self.goToPosition_button = QtWidgets.QPushButton(Form)
        self.goToPosition_button.setObjectName("goToPosition_button")
        self.gridLayout_2.addWidget(self.goToPosition_button, 13, 0, 1, 1)
        self.dr_label = QtWidgets.QLabel(Form)
        self.dr_label.setObjectName("dr_label")
        self.gridLayout_2.addWidget(self.dr_label, 9, 0, 1, 1)
        spacerItem = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.gridLayout_2.addItem(spacerItem, 14, 0, 1, 2)
        self.dX_spinbox = QtWidgets.QDoubleSpinBox(Form)
        self.dX_spinbox.setMinimum(-1000.0)
        self.dX_spinbox.setMaximum(1000.0)
        self.dX_spinbox.setObjectName("dX_spinbox")
        self.gridLayout_2.addWidget(self.dX_spinbox, 6, 1, 1, 1)
        self.moveRelative_button = QtWidgets.QPushButton(Form)
        self.moveRelative_button.setObjectName("moveRelative_button")
        self.gridLayout_2.addWidget(self.moveRelative_button, 11, 0, 1, 2)
        self.savedPositionName_lineEdit = QtWidgets.QLineEdit(Form)
        self.savedPositionName_lineEdit.setObjectName("savedPositionName_lineEdit")
        self.gridLayout_2.addWidget(self.savedPositionName_lineEdit, 12, 1, 1, 1)
        self.beam_type_combobox = QtWidgets.QComboBox(Form)
        self.beam_type_combobox.setObjectName("beam_type_combobox")
        self.beam_type_combobox.addItem("")
        self.beam_type_combobox.addItem("")
        self.gridLayout_2.addWidget(self.beam_type_combobox, 10, 1, 1, 1)
        self.dx_label = QtWidgets.QLabel(Form)
        self.dx_label.setObjectName("dx_label")
        self.gridLayout_2.addWidget(self.dx_label, 6, 0, 1, 1)
        self.dz_label = QtWidgets.QLabel(Form)
        self.dz_label.setObjectName("dz_label")
        self.gridLayout_2.addWidget(self.dz_label, 8, 0, 1, 1)
        self.insertManipulator_button = QtWidgets.QPushButton(Form)
        self.insertManipulator_button.setObjectName("insertManipulator_button")
        self.gridLayout_2.addWidget(self.insertManipulator_button, 3, 0, 1, 2)
        self.calibrated_status_label = QtWidgets.QLabel(Form)
        self.calibrated_status_label.setText("")
        self.calibrated_status_label.setObjectName("calibrated_status_label")
        self.gridLayout_2.addWidget(self.calibrated_status_label, 0, 0, 1, 1)
        self.beam_type_label = QtWidgets.QLabel(Form)
        self.beam_type_label.setObjectName("beam_type_label")
        self.gridLayout_2.addWidget(self.beam_type_label, 10, 0, 1, 1)
        self.savedPosition_combobox = QtWidgets.QComboBox(Form)
        self.savedPosition_combobox.setObjectName("savedPosition_combobox")
        self.gridLayout_2.addWidget(self.savedPosition_combobox, 13, 1, 1, 1)
        self.manipulatorStatus_label = QtWidgets.QLabel(Form)
        self.manipulatorStatus_label.setText("")
        self.manipulatorStatus_label.setObjectName("manipulatorStatus_label")
        self.gridLayout_2.addWidget(self.manipulatorStatus_label, 1, 0, 1, 2)
        self.dy_label = QtWidgets.QLabel(Form)
        self.dy_label.setObjectName("dy_label")
        self.gridLayout_2.addWidget(self.dy_label, 7, 0, 1, 1)
        self.move_type_comboBox = QtWidgets.QComboBox(Form)
        self.move_type_comboBox.setObjectName("move_type_comboBox")
        self.move_type_comboBox.addItem("")
        self.move_type_comboBox.addItem("")
        self.gridLayout_2.addWidget(self.move_type_comboBox, 5, 0, 1, 2)
        self.dR_spinbox = QtWidgets.QDoubleSpinBox(Form)
        self.dR_spinbox.setMinimum(-365.0)
        self.dR_spinbox.setMaximum(365.0)
        self.dR_spinbox.setObjectName("dR_spinbox")
        self.gridLayout_2.addWidget(self.dR_spinbox, 9, 1, 1, 1)
        self.addSavedPosition_button = QtWidgets.QPushButton(Form)
        self.addSavedPosition_button.setObjectName("addSavedPosition_button")
        self.gridLayout_2.addWidget(self.addSavedPosition_button, 12, 0, 1, 1)
        self.dY_spinbox = QtWidgets.QDoubleSpinBox(Form)
        self.dY_spinbox.setMinimum(-1000.0)
        self.dY_spinbox.setMaximum(1000.0)
        self.dY_spinbox.setObjectName("dY_spinbox")
        self.gridLayout_2.addWidget(self.dY_spinbox, 7, 1, 1, 1)
        self.pushButton_refresh_data = QtWidgets.QPushButton(Form)
        self.pushButton_refresh_data.setObjectName("pushButton_refresh_data")
        self.gridLayout_2.addWidget(self.pushButton_refresh_data, 4, 0, 1, 2)

        self.retranslateUi(Form)
        QtCore.QMetaObject.connectSlotsByName(Form)
        Form.setTabOrder(self.insertManipulator_button, self.move_type_comboBox)
        Form.setTabOrder(self.move_type_comboBox, self.dX_spinbox)
        Form.setTabOrder(self.dX_spinbox, self.dY_spinbox)
        Form.setTabOrder(self.dY_spinbox, self.dZ_spinbox)
        Form.setTabOrder(self.dZ_spinbox, self.dR_spinbox)
        Form.setTabOrder(self.dR_spinbox, self.beam_type_combobox)
        Form.setTabOrder(self.beam_type_combobox, self.moveRelative_button)
        Form.setTabOrder(self.moveRelative_button, self.addSavedPosition_button)
        Form.setTabOrder(self.addSavedPosition_button, self.savedPositionName_lineEdit)
        Form.setTabOrder(self.savedPositionName_lineEdit, self.goToPosition_button)
        Form.setTabOrder(self.goToPosition_button, self.savedPosition_combobox)

    def retranslateUi(self, Form):
        _translate = QtCore.QCoreApplication.translate
        Form.setWindowTitle(_translate("Form", "Form"))
        self.goToPosition_button.setText(_translate("Form", "Go To Position"))
        self.dr_label.setText(_translate("Form", "dR (deg)"))
        self.moveRelative_button.setText(_translate("Form", "Move "))
        self.beam_type_combobox.setItemText(0, _translate("Form", "ION"))
        self.beam_type_combobox.setItemText(1, _translate("Form", "ELECTRON"))
        self.dx_label.setText(_translate("Form", "dX (um)"))
        self.dz_label.setText(_translate("Form", "dZ (um)"))
        self.insertManipulator_button.setText(_translate("Form", "Insert"))
        self.beam_type_label.setText(_translate("Form", "Beam Type"))
        self.dy_label.setText(_translate("Form", "dY (um)"))
        self.move_type_comboBox.setItemText(0, _translate("Form", "Relative Move"))
        self.move_type_comboBox.setItemText(1, _translate("Form", "Corrected Move"))
        self.addSavedPosition_button.setText(_translate("Form", "Save Position"))
        self.pushButton_refresh_data.setText(_translate("Form", "Refresh Data"))



import numpy as np
import scipy.ndimage as ndi
from fibsem import conversions
from fibsem.imaging import utils as image_utils
from fibsem.structures import Point, FibsemImage
from PIL import Image, ImageDraw


### MASKING
def create_circle_mask(
    shape: tuple = (128, 128), radius: int = 32, sigma: int = 3
) -> np.ndarray:
    """_summary_

    Args:
        shape (tuple, optional): _description_. Defaults to (128, 128).
        radius (int, optional): _description_. Defaults to 32.
        sigma (int, optional): _description_. Defaults to 3.

    Returns:
        np.ndarray: _description_
    """
    distance = image_utils.create_distance_map_px(w=shape[1], h=shape[0])
    mask = distance <= radius

    if sigma:
        mask = ndi.filters.gaussian_filter(mask, sigma=sigma)

    return mask


def create_bandpass_mask(
    shape: tuple = (256, 256), lp: int = 32, hp: int = 2, sigma: int = 3
) -> np.ndarray:
    """_summary_

    Args:
        shape (tuple, optional): _description_. Defaults to (256, 256).
        lp (int, optional): _description_. Defaults to 32.
        hp (int, optional): _description_. Defaults to 2.
        sigma (int, optional): _description_. Defaults to 3.

    Returns:
        np.ndarray: _description_
    """

    distance = image_utils.create_distance_map_px(w=shape[1], h=shape[0])

    lowpass = distance <= lp
    highpass = distance >= hp

    mask = (lowpass * highpass).astype(np.float32)
    if sigma:
        mask = ndi.filters.gaussian_filter(mask, sigma=sigma)

    return mask


# FROM AUTOLAMELLA
def _mask_rectangular(image_shape, sigma=5.0, *, start=None, extent=None):
    """Make a rectangular mask with soft edges for image normalization.

    Parameters
    ----------
    image_shape : tuple
        Shape of the original image array
    sigma : float, optional
        Sigma value (in pixels) for gaussian blur function, by default 5.
    start : tuple, optional
        Origin point of the rectangle, e.g., ([plane,] row, column).
        Default start is 5% of the total image width and height.
    extent : int, optional
        The extent (size) of the drawn rectangle.
        E.g., ([num_planes,] num_rows, num_cols).
        Default is for the rectangle to cover 95% of the image width & height.

    Returns
    -------
    ndarray
        Rectangular mask with soft edges in array matching input image_shape.
    """
    import skimage

    if extent is None:
        # leave at least a 5% gap on each edge
        start = np.round(np.array(image_shape) * 0.05)
        extent = np.round(np.array(image_shape) * 0.90)
    rr, cc = skimage.draw.rectangle(start, extent=extent, shape=image_shape)
    mask = np.zeros(image_shape)
    mask[rr.astype(int), cc.astype(int)] = 1.0
    mask = ndi.gaussian_filter(mask, sigma=sigma)
    return mask


def create_rect_mask(
    img: np.ndarray, w: int, h: int, sigma: int = 0, pt: Point = None
) -> np.ndarray:
    """Create a rectangular mask at centred at the desired point.

    Args:
        img (np.ndarray): Image to be masked
        w (int): mask width
        h (int): mask height
        sigma (int, optional): gaussian blur to apply to mask (softness). Defaults to 0.
        pt (Point): Mask centre point. Defaults to None

    Returns:
        np.ndarray: mask
    """
    # centre mask point
    if pt is None:
        pt = Point(img.data.shape[1] // 2, img.data.shape[0] // 2)

    mask = np.zeros_like(img)

    y_min, y_max = int(np.clip(pt.y - h / 2, 0, img.shape[1])), int(
        np.clip(pt.y + h / 2, 0, img.shape[1])
    )
    x_min, x_max = int(np.clip(pt.x - w / 2, 0, img.shape[1])), int(
        np.clip(pt.x + w / 2, 0, img.shape[1])
    )

    mask[y_min:y_max, x_min:x_max] = 1

    if sigma:
        mask = ndi.filters.gaussian_filter(mask, sigma=sigma)

    return mask


def create_lamella_mask(
    img: FibsemImage,
    protocol: dict,
    scale: int = 2,
    circ: bool = False,
    pt: Point = None,
    use_trench_height: bool = False,
) -> np.ndarray:
    """Create a mask based on the size of the lamella

    Args:
        img (FibsemImage): reference image
        settings (dict): protocol dictionary
        scale (int, optional): mask size will be multipled by this scale . Defaults to 2.
        circ (bool, optional): use a circular mask. Defaults to False.
        pt (Point, optional): original point for the mask. Defaults to None.
        use_trench_height (bool, optional): use the trench height to calculate the mask size instead of lamella height). Defaults to False.

    Returns:
        np.ndarray: mask
    """

    # get the size of the lamella in pixels
    lamella_height_px, lamella_width_px = conversions.get_lamella_size_in_pixels(
        img, protocol, use_trench_height
    )

    if circ:
        mask = create_circle_mask(
            size=(img.data.shape[1], img.data.shape[0]),
            radius=max(lamella_height_px, lamella_width_px) * scale,
            sigma=12,
        )
    else:
        mask = create_rect_mask(
            img=img.data,
            pt=pt,
            w=int(lamella_width_px * scale),
            h=int(lamella_height_px * scale),
            sigma=3,
        )

    return mask


def apply_circular_mask(img: np.ndarray, radius: int, sigma: int = 0) -> np.ndarray:
    circ_mask = create_circle_mask(img.shape, radius=radius, sigma=sigma)

    if img.ndim == 3:
        circ_mask = np.moveaxis(np.array([circ_mask, circ_mask, circ_mask]), 0, 2)

    return img * circ_mask


def create_area_mask(arr, left=False, right=False, upper=False, lower=False):
    cy, cx = np.asarray(arr.shape) // 2

    # left
    left_mask = np.zeros_like(arr)
    left_mask[:, :cx] = 1

    # right
    right_mask = np.zeros_like(arr)
    right_mask[:, cx:] = 1

    # top
    top_mask = np.zeros_like(arr)
    top_mask[:cy, :] = 1

    # bottom
    bot_mask = np.zeros_like(arr)
    bot_mask[cy:, :] = 1

    if left is True:
        h_mask = left_mask
    elif right is True:
        h_mask = right_mask
    else:
        h_mask = np.ones_like(arr)

    if upper is True:
        v_mask = top_mask
    elif lower is True:
        v_mask = bot_mask
    else:
        v_mask = np.ones_like(arr)

    mask = np.logical_and((v_mask), (h_mask))

    return mask


def create_vertical_mask(arr, width=128):
    mask = np.zeros_like(arr)
    mid = arr.shape[1] // 2
    mask[:, mid - width : mid + width] = 1
    return mask



def create_mask(arr: np.ndarray, mask_info: dict) -> np.ndarray:

    # mask_info = {
    #     "type": "circle",
    #     "radius": 500,
    #     "sigma": 0,
    #     "invert": False
    # }

    if mask_info["type"] == "circle":
        mask = create_circle_mask(arr.shape, radius=mask_info["radius"], sigma=mask_info["sigma"])

    if mask_info["type"] == "rect":
        mask = create_rect_mask(arr, pt=mask_info["pt"], w=mask_info["w"], h=mask_info["h"], sigma=mask_info["sigma"])

    if mask_info["invert"]:
        mask = np.logical_not(mask)

    return mask


import logging
from fibsem.structures import (
    BeamType,
    FibsemImage,
    FibsemRectangle,
    FibsemMillingSettings,
    Point,
    ImageSettings,
    MicroscopeSettings
)
from fibsem.structures import (FibsemPatternSettings, FibsemRectangleSettings, 
                               FibsemCircleSettings, FibsemLineSettings, 
                               FibsemBitmapSettings)
from fibsem.patterning import FibsemMillingStage
from typing import Union
from fibsem.microscope import FibsemMicroscope


########################### SETUP


def setup_milling(
    microscope: FibsemMicroscope,
    mill_settings: FibsemMillingSettings = None,
):
    """Setup Microscope for Ion Beam Milling.

    Args:
        microscope (FibsemMicroscope): Fibsem microscope instance
        patterning_mode (str, optional): Ion beam milling patterning mode. Defaults to "Serial".
        hfw (float, optional): horizontal field width for milling. Defaults to 100e-6.
    """
    microscope.setup_milling(mill_settings = mill_settings)

def run_milling_drift_corrected(
    microscope: FibsemMicroscope, 
    milling_current: float,  
    image_settings: ImageSettings, 
    ref_image: FibsemImage, 
    reduced_area: FibsemRectangle = None,
) -> None:
    """Run Ion Beam Milling.

    Args:
        microscope (FibsemMicroscope): Fibsem microscope instance
        milling_current (float, optional): ion beam milling current. Defaults to None.
        asynch (bool, optional): flag to run milling asynchronously. Defaults to False.
    """
    microscope.run_milling_drift_corrected(milling_current, image_settings, ref_image, reduced_area)

def run_milling(
    microscope: FibsemMicroscope,
    milling_current: float,
    milling_voltage: float,
    asynch: bool = False,
) -> None:
    """Run Ion Beam Milling.

    Args:
        microscope (FibsemMicroscope): Fibsem microscope instance
        milling_current (float, optional): ion beam milling current. Defaults to None.
        asynch (bool, optional): flag to run milling asynchronously. Defaults to False.
    """
    microscope.run_milling(milling_current, milling_voltage, asynch)


def estimate_milling_time(microscope: FibsemMicroscope, microscope_patterns) -> float:
    """Get the milling status.

    Args:
        microscope (FibsemMicroscope): Fibsem microscope instance

    """

    total_time = microscope.estimate_milling_time(microscope_patterns)
        
    return total_time

def finish_milling(
    microscope: FibsemMicroscope, imaging_current: float = 20e-12, imaging_voltage: float = 30e3
) -> None:
    """Clear milling patterns, and restore to the imaging current.

    Args:
        microscope (FIbsemMicroscope): Fibsem microscope instance
        imaging_current (float, optional): Imaging Current. Defaults to 20e-12.

    """
    # restore imaging current
    logging.info(f"Changing to Imaging Current: {imaging_current:.2e}")
    microscope.finish_milling(imaging_current=imaging_current, imaging_voltage=imaging_voltage)
    logging.info("Finished Ion Beam Milling.")

def draw_patterns(microscope: FibsemMicroscope, patterns: list[FibsemPatternSettings]) -> None:
    """Draw a milling pattern from settings
    Args:
        microscope (FibsemMicroscope): Fibsem microscope instance
    """
    microscope_patterns = []
    for pattern in patterns:
        microscope_patterns.append(draw_pattern(microscope, pattern))
    return microscope_patterns

        
def draw_pattern(microscope: FibsemMicroscope, pattern: FibsemPatternSettings):
    """Draw a milling pattern from settings

    Args:
        microscope (FibsemMicroscope): Fibsem microscope instance
        pattern_settings (FibsemPatternSettings): pattern settings
        mill_settings (FibsemMillingSettings): milling settings
    """
    if isinstance(pattern, FibsemRectangleSettings):
        microscope_pattern = microscope.draw_rectangle(pattern)

    elif isinstance(pattern, FibsemLineSettings):
        microscope_pattern = microscope.draw_line(pattern)

    elif isinstance(pattern, FibsemCircleSettings):
        if pattern.thickness != 0:
            microscope_pattern = microscope.draw_annulus(pattern)
        else:
            microscope_pattern = microscope.draw_circle(pattern)

    elif isinstance(pattern, FibsemBitmapSettings):
        microscope_pattern = microscope.draw_bitmap_pattern(pattern, pattern.path)
        
    return microscope_pattern


def convert_to_bitmap_format(path):
    from PIL import Image
    import os 
    img=Image.open(path)
    a=img.convert("RGB", palette=Image.ADAPTIVE, colors=8)
    new_path = os.path.join(os.path.dirname(path), "24bit_img.tif")
    a.save(new_path)
    return new_path


def mill_stages(microscope: FibsemMicroscope, stages: list[FibsemMillingStage], asynch: bool=False):
    for stage in stages:
        mill_stage(microscope=microscope, stage=stage, asynch=asynch)

        # TODO: add special case for overtilt milling

def mill_stage(microscope: FibsemMicroscope, stage: FibsemMillingStage, asynch: bool=False):

    # set up milling
    setup_milling(microscope, stage.milling)

    # draw patterns
    for pattern in stage.pattern.patterns:
        draw_pattern(microscope, pattern)

    run_milling(microscope, stage.milling.milling_current, asynch)

    # finish milling
    finish_milling(microscope)



def mill_stages_ui(microscope: FibsemMicroscope, stages: list[FibsemMillingStage], asynch: bool=False, parent_ui = None):
    """Run a list of milling stages, with a progress bar and notifications."""
    import napari.utils.notifications

    if isinstance(stages, FibsemMillingStage):
        stages = [stages]

  
    for idx, stage in enumerate(stages):
    
        if parent_ui:
            parent_ui.milling_notification.emit(f"Preparing: {stage.name}")
    
        try:
            mill_stage_ui(microscope, stage, parent_ui, idx, len(stages))
            # TODO implement a special case for overtilt milling

            # drift correction

        except Exception as e:
            napari.utils.notifications.show_error(f"Error running milling stage: {stage.name}")
            logging.error(e)
        finally:
            finish_milling(microscope, 
                                        imaging_current=microscope.system.ion.beam.beam_current, 
                                        imaging_voltage=microscope.system.ion.beam.voltage)
        if parent_ui:
            parent_ui._progress_bar_quit.emit()
            parent_ui.milling_notification.emit(f"Milling stage complete: {stage.name}")
    
    if parent_ui:
        parent_ui.milling_notification.emit(f"Milling complete. {len(stages)} stages completed.")


def mill_stage_ui(microscope: FibsemMicroscope, stage: FibsemMillingStage, parent_ui = None, idx: int = 0, n_stages: int = 1):

    setup_milling(microscope, mill_settings=stage.milling)

    patterns = draw_patterns(microscope, stage.pattern.patterns)
    estimated_time = estimate_milling_time(microscope, patterns)
    logging.info(f"Estimated time for {stage.name}: {estimated_time:.2f} seconds")

    if parent_ui:
        progress_bar_dict = {"estimated_time": estimated_time, "idx": idx, "total": n_stages}
        parent_ui._progress_bar_start.emit(progress_bar_dict)
        parent_ui.milling_notification.emit(f"Running {stage.name}...")

    run_milling(microscope, stage.milling.milling_current, stage.milling.milling_voltage)

    return 



def mill_stage_with_overtilt(microscope: FibsemMicroscope, stage: FibsemMillingStage, ref_image: FibsemImage, asynch=False ):
    """Mill a trench pattern with overtilt, 
    based on https://www.sciencedirect.com/science/article/abs/pii/S1047847716301514 and autolamella v1"""
    # set up milling
    setup_milling(microscope, stage.milling)

    from fibsem.structures import FibsemStagePosition
    from fibsem import alignment, patterning
    import numpy as np

    overtilt_in_degrees = stage.pattern.protocol.get("overtilt", 1)

    # assert pattern is TrenchPattern
    if not isinstance(stage.pattern, patterning.TrenchPattern):
        raise ValueError("Pattern must be TrenchPattern for overtilt milling")

    for i, pattern in enumerate(stage.pattern.patterns):
        
        # save initial position
        initial_position = microscope.get_stage_position()
        
        # overtilt
        if i == 0:
            t = -np.deg2rad(overtilt_in_degrees)
        else:
            t = +np.deg2rad(overtilt_in_degrees)
        microscope.move_stage_relative(FibsemStagePosition(t=t))

        # beam alignment
        image_settings = ImageSettings.fromFibsemImage(ref_image)
        image_settings.filename = f"alignment_target_{stage.name}"
        
        alignment.multi_step_alignment_v2(microscope=microscope, 
                                        ref_image=ref_image, 
                                        beam_type=BeamType.ION, 
                                        alignment_current=None,
                                        steps=3)

        # setup again to ensure we are milling at the correct current, cleared patterns
        setup_milling(microscope, stage.milling)

        # draw pattern
        draw_pattern(microscope, pattern)

        # run milling
        run_milling(microscope, stage.milling.milling_current, asynch)

        # return to initial position
        microscope.move_stage_absolute(initial_position)

    # finish milling
    finish_milling(microscope)



############################# UTILS #############################



#!/usr/bin/env python3

import argparse
    
# import matplotlib.pyplot as plt
import numpy as np
import segmentation_models_pytorch as smp
from fibsem.segmentation import dataset, utils, validate_config
import torch
import os
import wandb
from PIL import Image
import yaml
import zarr
import glob
import tifffile as tff
import time

def inference(images, output_dir, model, model_path, device, WANDB=False):
    """Helper function for performing inference with the model"""
    # Load the model
    model.load_state_dict(torch.load(model_path))
    model = model.to(device)
    # model.eval()

    # Inference
    with torch.no_grad(): # TODO: move this down to only wrap the model inference
        vol = tff.imread(os.path.join(images, "*.tif*"), aszarr=True) # loading folder of .tif into zarr array)
        zarr_set = zarr.open(vol)

        filenames = sorted(glob.glob(os.path.join(images, "*.tif*")))

        for img, fname in zip(zarr_set, filenames):
            img = torch.tensor(np.asarray(img)).unsqueeze(0)
            img = img.to(device)
            outputs = model(img[None, :, :, :].float())
            output_mask = utils.decode_output(outputs)
            
            output = Image.fromarray(output_mask) 
            path = os.path.join(output_dir, os.path.basename(fname).split(".")[0])
            
            # if not os.path.exists(path):
            os.makedirs(path, exist_ok=True)
            
            input_img = Image.fromarray(img.detach().cpu().squeeze().numpy())
            input_img.save(os.path.join(path, "input.tif"))
            output.save(os.path.join(path, "output.tif"))  # or 'test.tif'

            if WANDB:
                img_base = img.detach().cpu().squeeze().numpy()
                img_rgb = np.dstack((img_base, img_base, img_base))

                wb_img = wandb.Image(img_rgb, caption="Input Image")
                wb_mask = wandb.Image(output_mask, caption="Output Mask")
                wandb.log({"image": wb_img, "mask": wb_mask})

        return outputs, output_mask

if __name__ == "__main__":
    # command line arguments
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--config",
        help="the directory containing the config file to use",
        dest="config",
        action="store",
        default=os.path.join("fibsem", "segmentation", "config.yml"),
    )
    args = parser.parse_args()
    config_dir = args.config

    # NOTE: Setup your config.yml file
    with open(config_dir, 'r') as f:
        config = yaml.safe_load(f)

    print("Validating config file.")
    validate_config.validate_config(config, "inference")

    # directories
    data_path = config["inference"]["data_dir"]
    model_weights = config["inference"]["model_dir"]
    output_dir = config["inference"]["output_dir"]

    # other parameters
    cuda = config["inference"]["cuda"]
    WANDB = config["inference"]["wandb"]


    model = smp.Unet(
            encoder_name=config["inference"]["encoder"],
            encoder_weights="imagenet",
            in_channels=1,  # grayscale images
            classes=config["inference"]["num_classes"],  # background, needle, lamella
        )

    if WANDB:
        # weights and biases setup
        wandb.init(project=config["inference"]["wandb_project"], entity=config["inference"]["wandb_entity"])

    device = torch.device("cuda:0" if torch.cuda.is_available() and cuda else "cpu")

    inference(data_path, output_dir, model, model_weights, device, WANDB)
    


import glob
import json
import os

import torch
import numpy as np
import matplotlib.pyplot as plt
from nnunetv2.imageio.simpleitk_reader_writer import SimpleITKIO
from nnunetv2.inference.predict_from_raw_data import nnUNetPredictor

import glob
import PIL.Image
from tqdm import tqdm
import os
import shutil

import json


def load_model(path: str, device) -> nnUNetPredictor:
    # instantiate the nnUNetPredictor
    model = nnUNetPredictor(
        tile_step_size=0.5,
        perform_everything_on_gpu=bool(device.type == "cuda"),
        device=device,
        verbose=False,
        verbose_preprocessing=False,
        allow_tqdm=False,
    )

    model = load_from_checkpoint(model, path)

    return model


def load_image(fname) -> np.ndarray:
    return SimpleITKIO().read_images([fname])[0]


def inference(model: nnUNetPredictor, image: np.ndarray):
    # todo: dont' return scores by defaults
    IMAGE_PROPERTIES = {
        "sitk_stuff": {
            "spacing": (1.0, 1.0),
            "origin": (0.0, 0.0),
            "direction": (1.0, 0.0, 0.0, 1.0),
        },
        "spacing": [999.0, 1.0, 1.0],
    }

    mask, scores = model.predict_single_npy_array(
        image, IMAGE_PROPERTIES, None, None, True
    )

    # mask = class map
    # scores = probability map for each class

    return mask, scores


def load_from_checkpoint(
    model: nnUNetPredictor, checkpoint_path: str
) -> nnUNetPredictor:
    """Load model from single checkpoint"""
    # TODO: this is a copy of nnunetv2.inference.predict_from_raw_data.load_from_checkpoint
    # should be depreciated once the change is upstreamed...

    from nnunetv2.utilities.plans_handling.plans_handler import PlansManager
    from nnunetv2.inference.predict_from_raw_data import determine_num_input_channels
    import nnunetv2
    from nnunetv2.utilities.find_class_by_name import recursive_find_python_class
    from nnunetv2.utilities.label_handling.label_handling import (
        determine_num_input_channels,
    )
    from nnunetv2.utilities.plans_handling.plans_handler import PlansManager
    from torch._dynamo import OptimizedModule

    # check if nnunetPredictor has method called load_from_checkpoint
    # TODO: remove once the change is upstreamed
    method_deprecated = hasattr(model, "load_from_checkpoint")
    if method_deprecated:
        print(f"this method is deprecated: {method_deprecated}")
        print("please update nnunetv2 to the latest version, and update...")

    # load the full checkpoint
    model_checkpoint = torch.load(checkpoint_path, map_location=torch.device("cpu"))

    # load the dataset and plans
    dataset_json = model_checkpoint["dataset"]
    plans_json = model_checkpoint["plans"]
    plans_manager = PlansManager(plans_json)

    # load the model parameters
    parameters = []
    checkpoint_name = "final" # always use final checkpoint for now
    for i, k in enumerate(sorted(model_checkpoint['folds'])):

        checkpoint = model_checkpoint['folds'][k][checkpoint_name]
        
        if i == 0: # use first fold to get trainer and configuration name
            trainer_name = checkpoint["trainer_name"]
            configuration_name = checkpoint["init_args"]["configuration"]
            inference_allowed_mirroring_axes = (
                checkpoint["inference_allowed_mirroring_axes"]
                if "inference_allowed_mirroring_axes" in checkpoint.keys()
                else None
            )

        parameters.append(checkpoint["network_weights"])

    configuration_manager = plans_manager.get_configuration(configuration_name)

    # restore network
    num_input_channels = determine_num_input_channels(
        plans_manager, configuration_manager, dataset_json
    )
    trainer_class = recursive_find_python_class(
        os.path.join(nnunetv2.__path__[0], "training", "nnUNetTrainer"),
        trainer_name,
        "nnunetv2.training.nnUNetTrainer",
    )
    network = trainer_class.build_network_architecture(
        plans_manager,
        dataset_json,
        configuration_manager,
        num_input_channels,
        enable_deep_supervision=False,
    )
    model.plans_manager = plans_manager
    model.configuration_manager = configuration_manager
    model.list_of_parameters = parameters
    model.network = network
    model.dataset_json = dataset_json
    model.trainer_name = trainer_name
    model.allowed_mirroring_axes = inference_allowed_mirroring_axes
    model.label_manager = plans_manager.get_label_manager(dataset_json)
    if (
        ("nnUNet_compile" in os.environ.keys())
        and (os.environ["nnUNet_compile"].lower() in ("true", "1", "t"))
        and not isinstance(model.network, OptimizedModule)
    ):
        print("Using torch.compile")
        model.network = torch.compile(model.network)

    return model


def export_model_checkpoint(path: str, checkpoint_path: str = None) -> None:
    """Save NNUNet model checkpoint as a single .pth file
    args:
        path: path to the nnunet model directory

    """
    # nnunet model directory structure for ensemble:
    # model
    #  dataset.json
    #  plans.json
    #  fold_n:
    #   checkpoint_best.pth
    #   checkpoint_final.pth

    # we want to convert it to a single .pth file with the following structure:
    # model_checkpoint.pth
    # dataset: dataset.json
    # plans: plans.json
    # fold_n:
    #  best:  checkpoint_best.pth
    #  final: checkpoint_final.pth

    # this makes it more portable and easier to load

    def load_json(path: str):
        with open(path, "r") as f:
            return json.load(f)

    MODEL_CHECKPOINT = {}

    # paths
    DATASET_JSON_PATH = os.path.join(path, "model", "dataset.json")
    PLAN_JSON_PATH = os.path.join(path, "model", "plans.json")

    # load the dataset and plans
    MODEL_CHECKPOINT["dataset"] = load_json(DATASET_JSON_PATH)
    MODEL_CHECKPOINT["plans"] = load_json(PLAN_JSON_PATH)

    # load the folds
    MODEL_CHECKPOINT["folds"] = {}

    # get all the fold directories,
    FOLD_DIRS = sorted(glob.glob(os.path.join(path, "model", "fold_*")))
    print(f"Found {len(FOLD_DIRS)} folds...")
    for fold_dir in FOLD_DIRS:
        fold_name = os.path.basename(fold_dir)
        print(f"Processing fold {fold_name}...")

        # load the best/ final checkpoint
        BEST_CHECKPOINT_PATH = os.path.join(fold_dir, "checkpoint_best.pth")
        FINAL_CHECKPOINT_PATH = os.path.join(fold_dir, "checkpoint_final.pth")

        MODEL_CHECKPOINT["folds"][fold_name] = {
            "best": torch.load(BEST_CHECKPOINT_PATH, map_location=torch.device("cpu")),
            "final": torch.load(
                FINAL_CHECKPOINT_PATH, map_location=torch.device("cpu")
            ),
        }

    # save as single torch checkpoint
    if checkpoint_path is None:
        checkpoint_path = os.path.join(path, "model_checkpoint.pth")
    torch.save(MODEL_CHECKPOINT, checkpoint_path)
    print(f"Saved model checkpoint to {checkpoint_path}")


def convert_to_nnunet_dataset(
    data_path, label_path, nnunet_data_path, label_map=None, filetype=".tif"
):
    """
    Convert a directory of images and labels to an nnunet dataset.
    """

    DATA_PATH = data_path
    LABEL_PATH = label_path
    LABEL_MAP = label_map
    NNUNET_DATA_PATH = nnunet_data_path
    FILETYPE = filetype

    # check if the NNUNET_DATA_PATH has a prefix of DATASET

    # check if is directory
    if os.path.exists(NNUNET_DATA_PATH) and not os.path.isdir(NNUNET_DATA_PATH):
        NNUNET_DATA_PATH = os.path.dirname(NNUNET_DATA_PATH)

    folder_basename = os.path.basename(NNUNET_DATA_PATH)

    if not folder_basename.startswith("Dataset"):
        # get the dirname of NNUNET_DATA_PATH
        dirname = os.path.dirname(NNUNET_DATA_PATH)
        # get all the folders in dirname
        n_folders = len(glob.glob(os.path.join(dirname, "Dataset*")))
        # get the next dataset number
        next_dataset_number = n_folders + 1

        # rename the folder
        new_nnunet_path = os.path.join(
            dirname, f"Dataset{next_dataset_number:03d}_{folder_basename}"
        )

        print(
            f"NNUnet requires the dataset folder to begin with DatasetXXX, where XXX is the number of the dataset."
        )
        ret = input(f"Rename {NNUNET_DATA_PATH} to {new_nnunet_path}? (y/n)")

        if ret.lower() == "y":
            NNUNET_DATA_PATH = new_nnunet_path

    NNUNET_IMAGES_PATH = os.path.join(NNUNET_DATA_PATH, "imagesTr")
    NNUNET_LABELS_PATH = os.path.join(NNUNET_DATA_PATH, "labelsTr")

    GLOB_PATTERN = f"*{FILETYPE}"
    SUPPORTED_PATTERNS = ["*.tif", "*.png"]

    if GLOB_PATTERN not in SUPPORTED_PATTERNS:
        raise ValueError(
            f"glob pattern {GLOB_PATTERN} not supported. Supported filetypes are: {SUPPORTED_PATTERNS}"
        )

    os.makedirs(NNUNET_IMAGES_PATH, exist_ok=True)
    os.makedirs(NNUNET_LABELS_PATH, exist_ok=True)

    paths = zip([DATA_PATH, LABEL_PATH], [NNUNET_IMAGES_PATH, NNUNET_LABELS_PATH])

    for path, out_path in paths:
        print(f"Copying files from {path}...")
        filenames = glob.glob(os.path.join(path, f"{GLOB_PATTERN}"))

        ret = input(f"Copy {len(filenames)} files to {out_path}? (y/n)")

        if ret == "y":
            for fname in tqdm(filenames):
                # copy file to new directory
                basename = os.path.basename(fname)

                # if this is the images, we need to prepend _0000 before the fileextension
                # because nnunet expects the images to be named like this??
                if "imagesTr" in out_path:
                    # check if it has _0000 already
                    if "_0000" not in basename:
                        basename = basename.replace(FILETYPE, f"_0000{FILETYPE}")

                new_fname = os.path.join(out_path, basename)
                shutil.copy(fname, new_fname)

    ret = input(f"Write dataset json to {NNUNET_DATA_PATH}? (y/n)")

    if ret.lower() == "y":
        # NOTE: could also use nnunetv2.dataset_conversion.generate_dataset_json.generate_dataset_json
        # but we have to get most of the info manually anyway...

        # load labels, get number of unique classes
        filenames = glob.glob(os.path.join(NNUNET_LABELS_PATH, f"{GLOB_PATTERN}"))
        dataset_json = {}
        dataset_json["channel_names"] = {"0": "map"}  # only support one channel for now
        dataset_json["numTraining"] = len(filenames)
        dataset_json["file_ending"] = FILETYPE

        # get unique labels
        unique_labels = set()
        for fname in tqdm(filenames):
            im = np.asarray(PIL.Image.open(fname))
            unique_labels.update(np.unique(im))

        # check if number of unique labels matches number of labels in LABEL_MAP
        if LABEL_MAP is not None:
            matched_labels = (len(unique_labels) == len(LABEL_MAP),)
            if not matched_labels:
                print(
                    f"WARNING: Number of unique labels ({len(unique_labels)}) does not match number of labels in LABEL_MAP ({len(LABEL_MAP)})"
                )
        else:
            print(f"WARNING: No label map provided. Using default label map.")
            LABEL_MAP = [f"label_{i:02d}" for i in range(len(unique_labels))]

        dataset_json["labels"] = {label: i for i, label in enumerate(LABEL_MAP)}

        DATASET_PATH = os.path.join(NNUNET_DATA_PATH, "dataset.json")
        print(f"Writing dataset json to: {DATASET_PATH}")

        with open(DATASET_PATH, "w") as f:
            json.dump(dataset_json, f, indent=4)

    # FINISHED
    n_images = len(glob.glob(os.path.join(NNUNET_IMAGES_PATH, f"{GLOB_PATTERN}")))
    n_labels = len(glob.glob(os.path.join(NNUNET_LABELS_PATH, f"{GLOB_PATTERN}")))

    print(f"-" * 50)
    print(f"Summary: ")
    print(f"{n_images} images in {NNUNET_IMAGES_PATH}")
    print(f"{n_labels} labels in {NNUNET_LABELS_PATH}")
    print(f"Number of unique labels: {len(unique_labels)}")
    print(f"Label Map used {LABEL_MAP}")
    print(f"FileType: {FILETYPE}")
    print(f"Dataset json written to: {DATASET_PATH}")

    # get last 3 chars, convert to int
    nnUnetRaw = os.path.dirname(os.path.dirname(NNUNET_DATA_PATH))
    DATASET_ID = int(os.path.basename(NNUNET_DATA_PATH)[7:10])  # DatasetXXX always

    print("-" * 50)
    print("To train nnUNet with this dataset:")
    print(f"Dataset ID: {DATASET_ID}")
    print(f"Dataset Path: {NNUNET_DATA_PATH}")

    print(f"\n1. Set the following environment variables:")
    print(f"nnUNet_raw={nnUnetRaw}")

    print(f"Currently, the environment variables are set to:")
    print(f"nnUNet_raw: {os.environ.get('nnUNet_raw')}")
    print(f"nnUNet_preprocessed: {os.environ.get('nnUNet_preprocessed')}")
    print(f"nnUNet_results: {os.environ.get('nnUNet_results')}")

    # pre-process the dataset
    print(f"\n2. Pre-Process the Dataset: ")
    print(f"nnUNetv2_plan_and_preprocess -d {DATASET_ID} --verify_dataset_integrity")

    # train the dataset
    print(f"\n3. Train the Model on the Dataset {DATASET_ID}: ")
    print(f"nnUNetv2_train {DATASET_ID} 2d all")

    print(f"Post Training: ")
    print(f"\nOnce training is complete, you can run inference on the dataset:")
    print(f"See doc/nnunet_inference.md for more details.")
    print(f"-" * 50)




import napari
import napari.utils.notifications
from PyQt5 import QtWidgets

from fibsem.detection.detection import DetectedFeatures
from fibsem.segmentation.model import load_model

from napari.qt.threading import thread_worker
from PyQt5.QtCore import pyqtSignal
from fibsem.ui import _stylesheets
from fibsem.ui.qtdesigner_files import FibsemSegmentationModelWidget
from fibsem.segmentation.model import SegmentationModel


CHECKPOINT_PATH = "autolamella-mega-20240107.pt"
SEGMENT_ANYTHING_AVAIABLE = False
try:
    SEGMENT_ANYTHING_AVAIABLE = True
    from fibsem.segmentation.sam_model import SamModelWrapper
except ImportError:
    pass


AVAILABLE_MODELS = ["SegmentationModel"]

if SEGMENT_ANYTHING_AVAIABLE:
    AVAILABLE_MODELS.append("SegmentAnythingModel")
RECOMMENDED_SAM_CHECKPOINTS = ["facebook/sam-vit-base", "Zigeng/SlimSAM-uniform-50"]

class FibsemSegmentationModelWidget(FibsemSegmentationModelWidget.Ui_Form, QtWidgets.QDialog):
    continue_signal = pyqtSignal(DetectedFeatures)
    model_loaded = pyqtSignal()

    def __init__(
        self,
        model: SegmentationModel = None,
        parent=None,
    ):
        super(FibsemSegmentationModelWidget, self).__init__(parent=parent)
        self.setupUi(self)

        self.model = model
        self.model_type = None
        self.setup_connections()

    def setup_connections(self):

        # model
        self.pushButton_load_model.clicked.connect(self.load_model)
        self.lineEdit_checkpoint.setText(CHECKPOINT_PATH)
        self.comboBox_model_type.addItems(AVAILABLE_MODELS)
        self.comboBox_model_type.currentIndexChanged.connect(self.update_model_type)

    def update_model_type(self):

        model_type = self.comboBox_model_type.currentText()

        if model_type == "SegmentationModel":
            self.lineEdit_checkpoint.setText(CHECKPOINT_PATH)
            self.lineEdit_checkpoint.setToolTip("Please use a pretrained model.")
            self.checkpoint_seg_button.setEnabled(True)
        elif model_type == "SegmentAnythingModel":

            self.lineEdit_checkpoint.setText(RECOMMENDED_SAM_CHECKPOINTS[0])
            self.lineEdit_checkpoint.setToolTip(f"""Any supported transformer SAM model from HuggingFace can be used. 
                                                \nRecommended: 
                                                \nLarge GPU: {RECOMMENDED_SAM_CHECKPOINTS[0]}
                                                \nSmall GPU / CPU: {RECOMMENDED_SAM_CHECKPOINTS[1]}""")
            self.checkpoint_seg_button.setEnabled(False)

    # thread this, as it can take a long time to download the models
    def load_model(self) -> SegmentationModel:

        model_type = self.comboBox_model_type.currentText()
        checkpoint = self.lineEdit_checkpoint.text()

        self.pushButton_load_model.setEnabled(False)
        self.pushButton_load_model.setText("Loading...")
        self.pushButton_load_model.setStyleSheet(_stylesheets._ORANGE_PUSHBUTTON_STYLE)
        self.pushButton_load_model.setToolTip("Downloading model... check terminal for progress...")

        worker = self.load_model_worker(model_type=model_type, checkpoint=checkpoint)
        worker.finished.connect(self.load_model_finished)
        worker.start()


    @thread_worker
    def load_model_worker(self, model_type: str, checkpoint: str):
        
        if model_type == "SegmentationModel":
            self.model = load_model(checkpoint=checkpoint)

        if model_type == "SegmentAnythingModel":
            self.model = SamModelWrapper(checkpoint=checkpoint)

        self.model_type = model_type
        self.model.checkpoint = checkpoint

        return self.model

    def load_model_finished(self):

        self.pushButton_load_model.setEnabled(True)
        self.pushButton_load_model.setText("Load Model")
        self.pushButton_load_model.setStyleSheet(_stylesheets._BLUE_PUSHBUTTON_STYLE)
        self.pushButton_load_model.setToolTip("")

        if self.model is not None:
            print(f"Loaded: {self.model}, {self.model.device}, {self.model.checkpoint}")
            self.model_loaded.emit()

def main():

    viewer  = napari.Viewer()
    widget = FibsemSegmentationModelWidget()
    viewer.window.add_dock_widget(widget, 
                                  area="right", 
                                  name="Fibsem Segmentation Model")

    napari.run()
    


if __name__ == "__main__":
    main()




import os


from fibsem.structures import (
    BeamType,
    ImageSettings,
    ReferenceImages,
    FibsemImage,
    FibsemRectangle,
)
from fibsem.microscope import FibsemMicroscope
from fibsem.imaging import autogamma


def new_image(
    microscope: FibsemMicroscope,
    settings: ImageSettings,
) -> FibsemImage:
    """Apply the given image settings and acquire a new image.

    Args:
        microscope (FibsemMicroscope): The FibsemMicroscope instance used to acquire the image.
        settings (ImageSettings): The image settings used to acquire the image.

    Returns:
        FibsemImage: The acquired image.
    """

    # set filename
    if settings.beam_type is BeamType.ELECTRON:
        filename = f"{settings.filename}_eb"

    if settings.beam_type is BeamType.ION:
        filename = f"{settings.filename}_ib"

    # run autocontrast
    if settings.autocontrast:
        microscope.autocontrast(beam_type=settings.beam_type)

    # acquire the image
    image = microscope.acquire_image(
        image_settings=settings,
    )

    if settings.autogamma:
        image = autogamma.auto_gamma(image, method="autogamma")

    # save image
    if settings.save:
        filename = os.path.join(settings.path, filename)
        image.save(path=filename)

    return image

def acquire_image(microscope:FibsemMicroscope, settings:ImageSettings) -> FibsemImage:
    """ passthrough for new_image to match internal api"""
    return new_image(microscope, settings)

def last_image(microscope: FibsemMicroscope, beam_type: BeamType) -> FibsemImage:
    """_summary_

    Args:
        microscope (FibsemMicroscope): microscope instance
        beam_type (BeamType): beam type for image

    Returns:
        FibsemImage: last image acquired by the microscope
    """
    return microscope.last_image(beam_type=beam_type)



def take_reference_images(
    microscope: FibsemMicroscope, image_settings: ImageSettings
) -> tuple[FibsemImage, FibsemImage]:
    """
    Acquires a pair of electron and ion reference images using the specified imaging settings and
    a FibsemMicroscope instance.

    Args:
        microscope (FibsemMicroscope): A FibsemMicroscope instance for imaging.
        image_settings (ImageSettings): An ImageSettings object with the desired imaging parameters.

    Returns:
        A tuple containing a pair of FibsemImage objects, representing the electron and ion reference
        images acquired using the specified microscope and image settings.

    """
    import time 
    from fibsem.microscope import TescanMicroscope

    tmp_beam_type = image_settings.beam_type
    
    # acquire electron image
    image_settings.beam_type = BeamType.ELECTRON
    eb_image = acquire_image(microscope, image_settings)
    
    # acquire ion image
    image_settings.beam_type = BeamType.ION
    if isinstance(microscope, TescanMicroscope):
        time.sleep(3)
    ib_image = acquire_image(microscope, image_settings)
    image_settings.beam_type = tmp_beam_type  # reset to original beam type

    return eb_image, ib_image
   

def take_set_of_reference_images(
    microscope: FibsemMicroscope,
    image_settings: ImageSettings,
    hfws: tuple[float],
    filename: str = "ref_image",
) -> ReferenceImages:
    """
    Takes a set of reference images at low and high magnification using a FibsemMicroscope.
    The image settings and half-field widths for the low- and high-resolution images are
    specified using an ImageSettings object and a tuple of two floats, respectively.
    The optional filename parameter can be used to customize the image labels.

    Args:
        microscope (FibsemMicroscope): A FibsemMicroscope object to acquire the images from.
        image_settings (ImageSettings): An ImageSettings object with the desired imaging parameters.
        hfws (Tuple[float, float]): A tuple of two floats specifying the horizontal field widths (in microns)
            for the low- and high-resolution images, respectively.
        filename (str, optional): A filename to be included in the image filenames. Defaults to "ref_image".

    Returns:
        A ReferenceImages object containing the low- and high-resolution electron and ion beam images.

    Notes:
        This function sets image_settings.save to True before taking the images.
        The returned ReferenceImages object contains the electron and ion beam images as FibsemImage objects.
    """
    # force save
    image_settings.save = True

    image_settings.hfw = hfws[0]
    image_settings.filename = f"{filename}_low_res"
    low_eb, low_ib = take_reference_images(microscope, image_settings)

    image_settings.hfw = hfws[1]
    image_settings.filename = f"{filename}_high_res"
    high_eb, high_ib = take_reference_images(microscope, image_settings)

    reference_images = ReferenceImages(low_eb, high_eb, low_ib, high_ib)


    # more flexible version
    # reference_images = []
    # for i, hfw in enumerate(hfws):
    #     image_settings.hfw = hfw
    #     image_settings.filename = f"{filename}_res_{i:02d}"
    #     eb_image, ib_image = take_reference_images(microscope, image_settings)
    #     reference_images.append([eb_image, ib_image])

    return reference_images



